#ifndef MAGMA_MIC_BINDINGS_H
#define MAGMA_MIC_BINDINGS_H

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <scif.h>
#include <mkl.h>
#include <map>

#include "magma.h"
#include "magma_z_bindings.h"
#include "magma_c_bindings.h"
#include "magma_d_bindings.h"
#include "magma_s_bindings.h"

#define START_OFFSET    0x80000
#define PAGE_SIZE       0x1000
#define VA_GEN_MIN      0x4000000000000000

#define PRINT_dA               50
#define ALLOCATE_MEM          100
#define FREE_MEM              101
#define CLOSE_MIC             200

#define MIC_FAILURE           303

#define READFROM_DEV          400
#define WRITETO_DEV           401

#define magma_mic_SGEMM       500
#define magma_mic_STRSM       501
#define magma_mic_SSYRK       502
#define magma_mic_SLARFT      503
#define magma_mic_SLARFB      504
#define magma_mic_STRANSPOSE  505
#define magma_mic_STRMM       506
#define magma_mic_SYNC        507
#define magma_mic_SLASWP      508
#define magma_mic_SGEMV       509

#define magma_mic_DGEMM       600
#define magma_mic_DTRSM       601
#define magma_mic_DSYRK       602
#define magma_mic_DLARFT      603
#define magma_mic_DLARFB      604
#define magma_mic_DTRANSPOSE  605
#define magma_mic_DTRMM       606
#define magma_mic_DYNC        607
#define magma_mic_DLASWP      608
#define magma_mic_DGEMV       609

#define magma_mic_CGEMM       700
#define magma_mic_CTRSM       701
#define magma_mic_CHERK       702
#define magma_mic_CLARFT      703
#define magma_mic_CLARFB      704
#define magma_mic_CTRANSPOSE  705
#define magma_mic_CTRMM       706
#define magma_mic_CYNC        707
#define magma_mic_CLASWP      708
#define magma_mic_CGEMV       709

#define magma_mic_ZGEMM       800
#define magma_mic_ZTRSM       801
#define magma_mic_ZHERK       802
#define magma_mic_ZLARFT      803
#define magma_mic_ZLARFB      804
#define magma_mic_ZTRANSPOSE  805
#define magma_mic_ZTRMM       806
#define magma_mic_ZYNC        807
#define magma_mic_ZLASWP      808
#define magma_mic_ZGEMV       809

// globals, defined in magma_mic_bindings.cpp
extern std::map<off_t, size_t> gDev_nbyte;
extern std::map<off_t, size_t> gHost_nbyte;
extern std::map<off_t, off_t> gDev_offset;
extern std::map<off_t, off_t> gHost_offset;
extern scif_epd_t gEpd;


typedef struct mic_memory_info {
    off_t address;
    off_t offset;
    size_t nbyte;
} magma_mic_mem_info;


typedef struct mic_print_param {
    int m;
    int n;
    off_t d_A;
    int lda;
} magma_mic_print_param;


void magma_mic_synchronize_kernels();

int magma_mic_fence_mark( magma_event_t *mark );

int magma_mic_fence_wait( magma_event_t mark );

int magma_mic_memcpy2D_async(
          void *dst, int dpitch,
    const void *src, int spitch,
    int width, int height, int kind,
    magma_queue_t queue, magma_event_t* event );

#endif // MAGMA_MIC_BINDINGS_H
