/*
 *   -- micMAGMA (version 0.1) --
 *      Univ. of Tennessee, Knoxville
 *      Univ. of California, Berkeley
 *      Univ. of Colorado, Denver
 *      April 2012
 *
 * @generated s Thu Dec 13 14:20:25 2012
 */

// defines stub functions, if no GPU implementation is available.
#include <stdio.h>
#include <stdlib.h>
#include "magma.h"

#ifdef HAVE_MIC

#include "magma_mic_bindings.h"


// ========================================
// copying sub-matrices (contiguous columns)
magma_err_t
magma_ssetmatrix(
    magma_int_t m, magma_int_t n,
    float const* hA_src, size_t hA_offset, magma_int_t ldha,
    magmaFloat_ptr    dA_dst, size_t dA_offset, magma_int_t ldda,
    magma_queue_t queue )
{
    magma_event_t event;
    magma_mic_synchronize_kernels();
    magma_mic_memcpy2D_async( dA_dst + dA_offset, ldda*sizeof(float),
                              hA_src + hA_offset, ldha*sizeof(float),
                              m*sizeof(float), n,
                              WRITETO_DEV, queue, &event );
    magma_mic_fence_wait( event );
    return MAGMA_SUCCESS;
}

// --------------------
magma_err_t
magma_sgetmatrix(
    magma_int_t m, magma_int_t n,
    magmaFloat_const_ptr dA_src, size_t dA_offset, magma_int_t ldda,
    float*          hA_dst, size_t hA_offset, magma_int_t ldha,
    magma_queue_t queue )
{
    magma_event_t event;
    magma_mic_synchronize_kernels();
    magma_mic_memcpy2D_async( hA_dst + hA_offset, ldha*sizeof(float),
                              dA_src + dA_offset, ldda*sizeof(float),
                              m*sizeof(float), n,
                              READFROM_DEV, queue, &event );
    magma_mic_fence_wait( event );
    return MAGMA_SUCCESS;
}

// --------------------
magma_err_t
magma_ssetmatrix_async(
    magma_int_t m, magma_int_t n,
    float const* hA_src, size_t hA_offset, magma_int_t ldha,
    magmaFloat_ptr    dA_dst, size_t dA_offset, magma_int_t ldda,
    magma_queue_t queue, magma_event_t *event )
{
    magma_mic_memcpy2D_async( dA_dst + dA_offset, ldda*sizeof(float),
                              hA_src + hA_offset, ldha*sizeof(float),
                              m*sizeof(float), n,
                              WRITETO_DEV, queue, event );
    return MAGMA_SUCCESS;
}

// --------------------
magma_err_t
magma_sgetmatrix_async(
    magma_int_t m, magma_int_t n,
    magmaFloat_const_ptr dA_src, size_t dA_offset, magma_int_t ldda,
    float*          hA_dst, size_t hA_offset, magma_int_t ldha,
    magma_queue_t queue, magma_event_t *event )
{
    magma_mic_memcpy2D_async( hA_dst + hA_offset, ldha*sizeof(float),
                              dA_src + dA_offset, ldda*sizeof(float),
                              m*sizeof(float), n,
                              READFROM_DEV, queue, event );
    return MAGMA_SUCCESS;
}

// --------------------
magma_err_t
magma_ssetintarray_async(
    magma_int_t lda, magma_int_t const* hA_src, size_t hA_offset, 
    magma_int_t *dA_dst, size_t dA_offset, magma_queue_t queue, magma_event_t *event )
{
	magma_mic_memcpy2D_async((void *)(dA_dst + dA_offset), lda * sizeof(int), (void *)(hA_src + hA_offset), lda * sizeof(int), lda * sizeof(int), 1, WRITETO_DEV, queue, event);
    return MAGMA_SUCCESS;
}




// ========================================
// BLAS functions
magma_err_t
magma_sgemm(
    magma_trans_t transA, magma_trans_t transB,
    magma_int_t m, magma_int_t n, magma_int_t k,
    float alpha, magmaFloat_const_ptr dA, size_t dA_offset, magma_int_t lda,
                              magmaFloat_const_ptr dB, size_t dB_offset, magma_int_t ldb,
    float beta,  magmaFloat_ptr       dC, size_t dC_offset, magma_int_t ldc,
    magma_queue_t handle )
{
    int err;
    magma_mic_sgemm_param gemm_param;

    gemm_param.transa = transA;
    gemm_param.transb = transB;
    gemm_param.m      = m;
    gemm_param.n      = n;
    gemm_param.k      = k;
    gemm_param.alpha  = alpha;
    gemm_param.a      = dA + dA_offset;
    gemm_param.lda    = lda;
    gemm_param.b      = dB + dB_offset;
    gemm_param.ldb    = ldb;
    gemm_param.beta   = beta;
    gemm_param.c      = dC + dC_offset;
    gemm_param.ldc    = ldc;
  
    int control_msg = magma_mic_SGEMM;

    if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
        err = errno;
        printf("scif_send failed with err %d\n", errno);
        fflush(stdout);
    }
    
    if ((err = scif_send(gEpd, &gemm_param, sizeof(gemm_param), 1)) <= 0) {
        err = errno;
        printf("scif_send failed with err %d\n", errno);
        fflush(stdout);
    }
    
    return MAGMA_SUCCESS;
}

// --------------------
magma_err_t
magma_sgemv(
    magma_trans_t transA,
    magma_int_t m, magma_int_t n,
    float alpha, magmaFloat_const_ptr dA, size_t dA_offset, magma_int_t lda,
                              magmaFloat_const_ptr dx, size_t dx_offset, magma_int_t incx,
    float beta,  magmaFloat_ptr       dy, size_t dy_offset, magma_int_t incy,
    magma_queue_t handle )
{
    int err;
    magma_mic_sgemv_param gemv_param;

    gemv_param.transa = transA;
    gemv_param.m      = m;
    gemv_param.n      = n;
    gemv_param.alpha  = alpha;
    gemv_param.a      = dA + dA_offset;
    gemv_param.lda    = lda;
    gemv_param.x      = dx + dx_offset;
    gemv_param.incx   = incx;
    gemv_param.beta   = beta;
    gemv_param.y      = dy + dy_offset;
    gemv_param.incy    = incy;

    int control_msg = magma_mic_SGEMV;
 
    if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }

    if ((err = scif_send(gEpd, &gemv_param, sizeof(gemv_param), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }

    return MAGMA_SUCCESS;
}

// --------------------
magma_err_t
magma_ssymm(
    magma_side_t side, magma_uplo_t uplo,
    magma_int_t m, magma_int_t n,
    float alpha, magmaFloat_const_ptr dA, size_t dA_offset, magma_int_t lda,
                              magmaFloat_const_ptr dB, size_t dB_offset, magma_int_t ldb,
    float beta,  magmaFloat_ptr       dC, size_t dC_offset, magma_int_t ldc,
    magma_queue_t handle )
{
    printf( "%s( %d, %d, %d, %d,    %f, ptr, %ld, %d, ptr, %ld, %d, %f, ptr, %ld, %d, handle )\n",
        __func__,
        side, uplo, m, n,
        MAGMA_S_REAL(alpha), dA_offset, lda,
                             dB_offset, ldb,
        MAGMA_S_REAL(beta),  dC_offset, ldc );
    return 0;
}

// --------------------
magma_err_t
magma_ssymv(
    magma_uplo_t uplo,
    magma_int_t n,
    float alpha, magmaFloat_const_ptr dA, size_t dA_offset, magma_int_t lda,
                              magmaFloat_const_ptr dx, size_t dx_offset, magma_int_t incx,
    float beta,  magmaFloat_ptr       dy, size_t dy_offset, magma_int_t incy,
    magma_queue_t handle )
{
    printf( "%s(      %d, %d,       %f, ptr, %ld, %d, ptr, %ld, %d, %f, ptr, %ld, %d, handle )\n",
        __func__,
        uplo, n,
        MAGMA_S_REAL(alpha), dA_offset, lda,
                             dx_offset, incx,
        MAGMA_S_REAL(beta),  dy_offset, incy );
    return 0;
}

// --------------------
magma_err_t
magma_ssyrk(
    magma_uplo_t uplo,
    magma_trans_t trans,
    magma_int_t n, magma_int_t k,
    float alpha, magmaFloat_const_ptr dA, size_t dA_offset, magma_int_t lda,
    float beta,  magmaFloat_ptr       dC, size_t dC_offset, magma_int_t ldc,
    magma_queue_t handle )
{
    int err;
    magma_mic_ssyrk_param syrk_param;

    syrk_param.uplo  = uplo;
    syrk_param.trans = trans;
    syrk_param.n     = n;
    syrk_param.k     = k;
    syrk_param.alpha = alpha;
    syrk_param.a     = dA + dA_offset;
    syrk_param.lda   = lda;
    syrk_param.beta  = beta;
    syrk_param.c     = dC + dC_offset;
    syrk_param.ldc   = ldc;

    int control_msg = magma_mic_SSYRK;

    if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }
    
    if ((err = scif_send(gEpd, &syrk_param, sizeof(syrk_param), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }
    
    return MAGMA_SUCCESS;
}

// --------------------
magma_err_t
magma_strsm(
    magma_side_t  side,
    magma_uplo_t  uplo,
    magma_trans_t trans,
    magma_diag_t  diag,
    magma_int_t m, magma_int_t n,
    float alpha, magmaFloat_const_ptr dA, size_t dA_offset, magma_int_t lda,
                              magmaFloat_ptr       dB, size_t dB_offset, magma_int_t ldb,
    magma_queue_t handle )
{
    int err;
    magma_mic_strsm_param trsm_param;

    trsm_param.side   = side;
    trsm_param.uplo   = uplo;
    trsm_param.transa = trans;
    trsm_param.diag   = diag;
    trsm_param.m      = m;
    trsm_param.n      = n;
    trsm_param.alpha  = alpha;
    trsm_param.a      = dA + dA_offset;
    trsm_param.lda    = lda;
    trsm_param.b      = dB + dB_offset;
    trsm_param.ldb    = ldb;

    int control_msg = magma_mic_STRSM;

    if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }
    
    if ((err = scif_send(gEpd, &trsm_param, sizeof(trsm_param), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }
    
    return MAGMA_SUCCESS;
}

// --------------------                                                                                                                                      
magma_err_t
magma_strmm(
    magma_side_t side, magma_uplo_t uplo, magma_trans_t trans, magma_diag_t diag,
    magma_int_t m, magma_int_t n,
    float alpha, magmaFloat_const_ptr dA, size_t dA_offset, magma_int_t lda,
                              magmaFloat_ptr       dB, size_t dB_offset, magma_int_t ldb,
    magma_queue_t queue )
{
    int err;
    magma_mic_strmm_param trmm_param;

    trmm_param.side   = side;
    trmm_param.uplo   = uplo;
    trmm_param.transa = trans;
    trmm_param.diag   = diag;
    trmm_param.m      = m;
    trmm_param.n      = n;
    trmm_param.alpha  = alpha;
    trmm_param.a      = dA + dA_offset;
    trmm_param.lda    = lda;
    trmm_param.b      = dB + dB_offset;
    trmm_param.ldb    = ldb;

    int control_msg = magma_mic_STRMM;

    if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }
    
    if ((err = scif_send(gEpd, &trmm_param, sizeof(trmm_param), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }
    
    return MAGMA_SUCCESS;
}

// ========================================
// Auxuliary routines
magma_err_t
magma_stranspose(magmaFloat_ptr odata, int offo, int ldo,
                 magmaFloat_ptr idata, int offi, int ldi,
                 int m, int n,
                 magma_queue_t queue) 
{
    int err;
    magma_mic_stranspose_param transpose_param;

    transpose_param.odata = odata + offo;
    transpose_param.ldo   = ldo;
    transpose_param.idata = idata + offi;
    transpose_param.ldi   = ldi;
    transpose_param.m     = m;
    transpose_param.n     = n;

    int control_msg = magma_mic_STRANSPOSE;

    if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }

    if ((err = scif_send(gEpd, &transpose_param, sizeof(transpose_param), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }
    
    return MAGMA_SUCCESS;
}

/*magma_err_t
magma_slaswp(magmaFloat_ptr d_AT, magma_int_t lda, magma_int_t *d_ipiv, magma_int_t *h_ipiv, magma_int_t nb, magma_int_t ind, magma_queue_t queue)
{
	magma_event_t event;
	int err;
	magma_mic_slaswp_param slaswp_param;
	slaswp_param.d_AT		=	d_AT;
	slaswp_param.lda		=	lda;
	slaswp_param.nb			=	nb;
	slaswp_param.ind		=	ind;
	slaswp_param.d_ipiv		= (int *)d_ipiv;

	int control_msg		=	magma_mic_SLASWP;
	if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
		err = errno;
		printf("scif_send failed with err %d\n", errno);
		fflush(stdout);	
	}
	
	if ((err = scif_send(gEpd, &slaswp_param, sizeof(slaswp_param), 1)) <= 0) {
		err = errno;
		printf("scif_send failed with err %d\n", errno);
		fflush(stdout);	
	}
	return MAGMA_SUCCESS;
}*/

magma_err_t
magma_slaswp(magma_int_t n, magmaFloat_ptr d_A, magma_int_t lda, magma_int_t k1, magma_int_t k2, magma_int_t *d_ipiv, magma_int_t incx, magma_queue_t queue)
{
	magma_event_t event;
	int err;
	magma_mic_slaswp_param slaswp_param;
	slaswp_param.n			=	n;
	slaswp_param.d_A		=	d_A;
	slaswp_param.lda		=	lda;
	slaswp_param.k1			=	k1;
	slaswp_param.k2			=	k2;
	slaswp_param.d_ipiv		= 	d_ipiv;
	slaswp_param.incx		=	incx;

	int control_msg		=	magma_mic_SLASWP;
	if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
		err = errno;
		printf("scif_send failed with err %d\n", errno);
		fflush(stdout);	
	}
	
	if ((err = scif_send(gEpd, &slaswp_param, sizeof(slaswp_param), 1)) <= 0) {
		err = errno;
		printf("scif_send failed with err %d\n", errno);
		fflush(stdout);	
	}
	return MAGMA_SUCCESS;
}



// This version sends a larfb request to the MIC. The difference with the version in
// 'src' is that here there is only one request and control is passed immediately back
// to the CPU. 
magma_err_t
magma_slarfb( int side, int trans, int direct, int storev,
              magma_int_t m, magma_int_t n, magma_int_t k,
              magmaFloat_ptr dV, size_t dV_offset, magma_int_t ldv,
              magmaFloat_ptr dT, size_t dT_offset, magma_int_t ldt,
              magmaFloat_ptr dC, size_t dC_offset, magma_int_t ldc,
              magmaFloat_ptr dwork, size_t dwork_offset, magma_int_t ldwork,
              magma_queue_t queue)
{
    int err;
    magma_mic_slarfb_param larfb_param;
  
    larfb_param.side   = side;
    larfb_param.trans  = trans;
    larfb_param.direct = direct;
    larfb_param.storev = storev;
    larfb_param.m      = m;
    larfb_param.n      = n;
    larfb_param.k      = k;
    larfb_param.v      = dV + dV_offset;
    larfb_param.ldv    = ldv;
    larfb_param.t      = dT + dT_offset;
    larfb_param.ldt    = ldt;
    larfb_param.c      = dC + dC_offset;
    larfb_param.ldc    = ldc;
    larfb_param.work   = dwork + dwork_offset;
    larfb_param.ldwork = ldwork;

    int control_msg = magma_mic_SLARFB;

    if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }
    
    if ((err = scif_send(gEpd, &larfb_param, sizeof(larfb_param), 1)) <= 0) {
      err = errno;
      printf("scif_send failed with err %d\n", errno);
      fflush(stdout);
    }
    
    return MAGMA_SUCCESS;
}

#endif // HAVE_MIC
