/*
 *   -- micMAGMA (version 0.1) --
 *      Univ. of Tennessee, Knoxville
 *      Univ. of California, Berkeley
 *      Univ. of Colorado, Denver
 *      April 2012
 *
 * @generated c Thu Dec 13 14:39:22 2012
 */

#include <mkl.h>

#include <mkl_trans.h>
#include <scif.h>
#include <fcntl.h>
#include <offload.h>

//#ifdef __MIC__
//#define _M512_TYPED
//#include <lmmintrin.h>
//#endif
#include <immintrin.h>


#include "magma.h"
//#include "magma_c_bindings.h"
#include "magma_c_server.h"



void call_cgemm(magma_mic_cgemm_param cgemm_param) {
	magma_trans_t transa, transb;
	magma_int_t m, n, k, lda, ldb, ldc;
	magmaFloatComplex alpha, beta, *a, *b, *c;
	
        transa = cgemm_param.transa;
        transb = cgemm_param.transb;
        m = cgemm_param.m;
        n = cgemm_param.n;
        k = cgemm_param.k;
        alpha = cgemm_param.alpha;
        a = (magmaFloatComplex *)cgemm_param.a;
        lda = cgemm_param.lda;
        b = (magmaFloatComplex *)cgemm_param.b;
        ldb = cgemm_param.ldb;
        beta = cgemm_param.beta;
        c = (magmaFloatComplex *)cgemm_param.c;
        ldc = cgemm_param.ldc;

        //printf("CGEMM: %c %c %d %d %d %f 0x%lx %d 0x%lx %d %f 0x%lx %d\n", transa, transb, m, n, k, alpha, (off_t)a, lda, (off_t)b, ldb, beta, (off_t)c, ldc);

	cgemm( lapack_const(transa), lapack_const(transb), 
	       &m, &n, &k, (const MKL_Complex8 *)&alpha, (MKL_Complex8 *)a, &lda, 
	       (MKL_Complex8 *)b, &ldb, (const MKL_Complex8 *)&beta, 
	       (MKL_Complex8 *)c, &ldc);

	return;
}

void call_cgemv(magma_mic_cgemv_param cgemv_param) {
  magma_trans_t transa;
  magma_int_t m, n, lda, incx, incy;
  magmaFloatComplex alpha, beta, *a, *x, *y;

  transa = cgemv_param.transa;
  m      = cgemv_param.m;
  n      = cgemv_param.n;
  alpha  = cgemv_param.alpha;
  a      = (magmaFloatComplex *)cgemv_param.a;
  lda    = cgemv_param.lda;
  x      = (magmaFloatComplex *)cgemv_param.x;
  incx   = cgemv_param.incx;
  beta   = cgemv_param.beta;
  y      = (magmaFloatComplex *)cgemv_param.y;
  incy   = cgemv_param.incy;

  //printf("CGEMV: %c %c %d %d %d %f 0x%lx %d 0x%lx %d %f 0x%lx %d\n", transa, transb, m, n, k, alpha, (off_t)a, lda, (off_t)b, ldb, beta, (off_t)c, ldc);                                                                           

  cgemv( lapack_const(transa), &m, &n, (const MKL_Complex8 *)&alpha, 
	 (MKL_Complex8 *)a, &lda, (MKL_Complex8 *)x, &incx, 
	 (const MKL_Complex8 *)&beta, (MKL_Complex8 *)y, &incy);

  return;
}

void call_ctrsm(magma_mic_ctrsm_param ctrsm_param) {
        magma_side_t side;
	magma_uplo_t uplo;
	magma_trans_t transa;
	magma_diag_t diag;
	magma_int_t m, n, lda, ldb;
	magmaFloatComplex alpha, *a, *b;
	
        side = ctrsm_param.side;
        uplo = ctrsm_param.uplo;
        transa = ctrsm_param.transa;
        diag = ctrsm_param.diag;
        m = ctrsm_param.m;
        n = ctrsm_param.n;
        alpha = ctrsm_param.alpha;
        a = (magmaFloatComplex *)ctrsm_param.a;
        lda = ctrsm_param.lda;
        b = (magmaFloatComplex *)ctrsm_param.b;
        ldb = ctrsm_param.ldb;

        //printf("CTRSM: %c %c %c %c %d %d %f 0x%lx %d 0x%lx %d\n", side, uplo, transa, diag, m, n, alpha, (off_t)a, lda, (off_t)b, ldb);
	ctrsm( lapack_const(side), lapack_const(uplo), lapack_const(transa), lapack_const(diag),
	       &m, &n, (const MKL_Complex8 *)&alpha, (MKL_Complex8 *)a, &lda, 
	       (MKL_Complex8 *)b, &ldb);

	return;
}


void call_ctrmm(magma_mic_ctrmm_param ctrmm_param) {
        magma_side_t side;
	magma_uplo_t uplo;
	magma_trans_t transa;
	magma_diag_t diag;
        magma_int_t m, n, lda, ldb;
        magmaFloatComplex alpha, *a, *b;

        side = ctrmm_param.side;
        uplo = ctrmm_param.uplo;
        transa = ctrmm_param.transa;
        diag = ctrmm_param.diag;
        m = ctrmm_param.m;
        n = ctrmm_param.n;
        alpha = ctrmm_param.alpha;
        a = (magmaFloatComplex *)ctrmm_param.a;
        lda = ctrmm_param.lda;
        b = (magmaFloatComplex *)ctrmm_param.b;
        ldb = ctrmm_param.ldb;

        //printf("CTRMM: %c %c %c %c %d %d %f 0x%lx %d 0x%lx %d\n", side, uplo, transa, diag, m, n, alpha, (off_t)a, lda, (off_t)b, ldb);
	ctrmm(lapack_const(side), lapack_const(uplo), lapack_const(transa), lapack_const(diag),
	      &m, &n, (const MKL_Complex8 *)&alpha, (MKL_Complex8 *)a, &lda, 
	      (MKL_Complex8 *)b, &ldb);

	return;
}
        


void call_cherk(magma_mic_cherk_param cherk_param) {
        magma_uplo_t uplo;
	magma_trans_t trans;
	magma_int_t n, k, lda, ldc;
	float alpha, beta;
	magmaFloatComplex *a, *c;
	
        uplo = cherk_param.uplo;
        trans = cherk_param.trans;
        n = cherk_param.n;
        k = cherk_param.k;
        alpha = cherk_param.alpha;
        a = (magmaFloatComplex *)cherk_param.a;
        lda = cherk_param.lda;
        beta = cherk_param.beta;
        c = (magmaFloatComplex *)cherk_param.c;
        ldc = cherk_param.ldc;

        //printf("CHERK: %c %c %d %d %f 0x%lx %d %f 0x%lx %d\n", uplo, trans, n, k, alpha, (off_t)a, lda, beta, (off_t)c, ldc);

	cherk(lapack_const(uplo), lapack_const(trans), 
	      &n, &k, &alpha, (MKL_Complex8 *)a, &lda, 
	      &beta, (MKL_Complex8 *)c, &ldc);

	return;
}


void call_clarft(magma_mic_clarft_param clarft_param) {
	int direct, storev;
	magma_int_t n, k, ldv, ldt; 
	magmaFloatComplex *v, *tau, *t;
 
        direct = clarft_param.direct;
        storev = clarft_param.storev;
        n = clarft_param.n;
        k = clarft_param.k;
        v = (magmaFloatComplex *)clarft_param.v;
        ldv = clarft_param.ldv;
        tau = (magmaFloatComplex *)clarft_param.tau;
        t = (magmaFloatComplex *)clarft_param.t;
        ldt = clarft_param.ldt;

	//printf("ZLARFT: %c %c %d %d 0x%lx %d ox%lx 0x%lx %d\n", direct, storev, n, k, (off_t)v, ldv, (off_t)tau, (off_t)t, ldt);

	clarft(lapack_const(direct), lapack_const(storev),
	       &n, &k, (MKL_Complex8 *)v, &ldv, 
	       (MKL_Complex8 *)tau, (MKL_Complex8 *)t, &ldt);
	
	return;
}


void call_clarfb(magma_mic_clarfb_param clarfb_param) {
	int side, trans, direct, storev;
	magma_int_t m, n, k, ldv, ldt, ldc, ldwork;
	magmaFloatComplex *v, *t, *c, *work;

        side = clarfb_param.side;
        trans = clarfb_param.trans;
        direct = clarfb_param.direct;
        storev = clarfb_param.storev;
        m = clarfb_param.m;
        n = clarfb_param.n;
        k = clarfb_param.k;
        v = (magmaFloatComplex *)clarfb_param.v;
        ldv = clarfb_param.ldv;
        t = (magmaFloatComplex *)clarfb_param.t;
        ldt = clarfb_param.ldt;
        c = (magmaFloatComplex *)clarfb_param.c;
        ldc = clarfb_param.ldc;
        work = (magmaFloatComplex *)clarfb_param.work;
        ldwork = clarfb_param.ldwork;

        //printf("CLARFB: %c %c %c %c %d %d %d 0x%lx %d ox%lx %d 0x%lx %d 0x%lx %d\n", side, trans, direct, storev, m, n, k, (off_t)v, ldv, (off_t)t, ldt, (off_t)c, ldc, (off_t)work, ldwork); 
	clarfb(lapack_const(side), lapack_const(trans), lapack_const(direct), lapack_const(storev),
			&m, &n, &k, (MKL_Complex8 *)v, &ldv, (MKL_Complex8 *)t, &ldt, 
			(MKL_Complex8 *)c, &ldc, (MKL_Complex8 *)work, &ldwork);

	/*
	magmaFloatComplex c_zero    = MAGMA_C_MAKE(  0.0, 0.0 );
	magmaFloatComplex c_one     = MAGMA_C_MAKE(  1.0, 0.0 );
	magmaFloatComplex c_neg_one = MAGMA_C_MAKE( -1.0, 0.0 );

	if (m <= 0 || n <= 0) {
		return;
	}

	magma_int_t transt;
	if (trans == MagmaNoTrans) 
		transt = MagmaConjTrans;
	else
		transt = MagmaNoTrans;

	if ( side  == MagmaLeft ) {

		if ( storev == MagmaColumnwise ) 
		{
			cgemm( lapack_const(MagmaConjTrans), lapack_const(MagmaNoTrans),
					&n, &k, &m,
					&c_one,  (MKL_Complex8 *)c,    &ldc,
					(MKL_Complex8 *)v,    &ldv,
					&c_zero, (MKL_Complex8 *)work, &ldwork);

			if (direct == MagmaForward)
				ctrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
						&n, &k, 
						&c_one, (MKL_Complex8 *)t,    &ldt, 
						(MKL_Complex8 *)work, &ldwork);
			else
				ctrmm( lapack_const(MagmaRight), lapack_const(MagmaLower), lapack_const(transt), lapack_const(MagmaNonUnit),
						&n, &k, 
						&c_one, (MKL_Complex8 *)t,    &ldt, 
						(MKL_Complex8 *)work, &ldwork);

			cgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaConjTrans), 
					&m, &n, &k, 
					&c_neg_one, (MKL_Complex8 *)v,    &ldv,
					(MKL_Complex8 *)work, &ldwork, 
					&c_one,     (MKL_Complex8 *)c,    &ldc);
		}
		else {
			cgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaConjTrans), 
					&m, &k, &n, 
					&c_one,  (MKL_Complex8 *)c,    &ldc,
					(MKL_Complex8 *)v,    &ldv, 
					&c_zero, (MKL_Complex8 *)work, &ldwork);

			ctrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
					&m, &k, 
					&c_one, (MKL_Complex8 *)t,    &ldt, 
					(MKL_Complex8 *)work, &ldwork);

			cgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans), 
					&m, &n, &k, 
					&c_neg_one, (MKL_Complex8 *)work, &ldwork,
					(MKL_Complex8 *)v,    &ldv,
					&c_one,     (MKL_Complex8 *)c,    &ldc);
		}
	}

	else {

		if ( storev == MagmaColumnwise ) {
			cgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans),
					&m, &k, &n,
					&c_one,  (MKL_Complex8 *)c,    &ldc,
					(MKL_Complex8 *)v,    &ldv,
					&c_zero, (MKL_Complex8 *)work, &ldwork);
			// ??? ldwork replaced by k for case n < k

			if (direct == MagmaForward)
				ctrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
						&m, &k,
						&c_one, (MKL_Complex8 *)t,    &ldt,
						(MKL_Complex8 *)work, &ldwork);
			else
				ctrmm( lapack_const(MagmaRight), lapack_const(MagmaLower), lapack_const(transt), lapack_const(MagmaNonUnit),
						&m, &k,
						&c_one, (MKL_Complex8 *)t,    &ldt,
						(MKL_Complex8 *)work, &ldwork);

			cgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaConjTrans),
					&m, &n, &k,
					&c_neg_one, (MKL_Complex8 *)work, &ldwork, 
					(MKL_Complex8 *)v,    &ldv,
					&c_one,     (MKL_Complex8 *)c,    &ldc);
		}
		else {
			cgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaConjTrans),
					&m, &k, &n,
					&c_one,  (MKL_Complex8 *)c,    &ldc,
					(MKL_Complex8 *)v,    &ldv,
					&c_zero, (MKL_Complex8 *)work, &ldwork);

			ctrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
					&m, &k,
					&c_one, (MKL_Complex8 *)t,    &ldt,
					(MKL_Complex8 *)work, &ldwork);

			cgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans),
					&m, &n, &k,
					&c_neg_one, (MKL_Complex8 *)work, &ldwork,
					(MKL_Complex8 *)v,    &ldv,
					&c_one,     (MKL_Complex8 *)c,    &ldc);
		}
	}*/

	return;
}


void call_ctranspose(magma_mic_ctranspose_param ctranspose_param)
{
	/******************************************************
	 * _mm512_load_pd -- for loading
	 * _mm512_i32lo[ext]scatter_pd
	 * _mm512_i32loscatter_pd is the simplified version of ...ext...
	 * they do the same thing, the latter needs less arguments.
	 * **************************************************/
	//	printf("entered stranpose\n");
	//	fflush(stdout);
	float *odata, *idata;
	int ldo,ldi, m, n;
	int num_threads;
	int tid;
	int index[16], i;
	__M512C _reg;
	__m512i _index;
	float *my_head, *loop_head, *dest_head;
	int num_cols, num_cols_g, col, iter, num_iter, num_rows_g, num_rows;

	odata	=	(float *) ctranspose_param.odata;
	ldo		=	ctranspose_param.ldo;
	idata	=	(float *) ctranspose_param.idata;
	ldi		=	ctranspose_param.ldi;
	m		=	ctranspose_param.m;
	n		=	ctranspose_param.n;

	if (m == 0 || n == 0) {
		return;
	}
	if (m <= n) {
		for (i = 0; i < 16; i++) {
			index[i] = i*CREG_WIDTH*ldo;
		}
		// 62 cores, 4 hardware threads per core. Leave out one core for the OS
		// (62 - 1)*4 = 244

		num_threads = 244;
		num_cols_g =	(n + num_threads - 1)/num_threads; 
		num_iter = (int) m*1.0/CREG_WIDTH;
		_index = _mm512_set_epi32(index[15], index[14], index[13], index[12],
				index[11], index[10], index[9], index[8], index[7], index[6],
				index[5], index[4], index[3], index[2], index[1], index[0]);

#pragma omp parallel num_threads(num_threads) private(tid, col, num_cols, iter,\
		my_head, loop_head, dest_head, _reg)
		{
			tid = omp_get_thread_num();
			num_cols = num_cols_g;
			if (tid == (num_threads - 1) && (n%num_cols_g != 0)) { 
				num_cols = n%num_cols_g;
			}
			// starting point of my section, constant for each thread
			my_head = idata + ldi*(tid*num_cols_g);
			for (iter = 0; iter < num_iter; iter++) { // loop inside each column
				for (col = 0; col < num_cols; col++) { // loop over each column
					if (tid*num_cols_g + col < n) {
						// starting point of the inner loop
						loop_head = my_head + col*ldi + iter*CREG_WIDTH;
						dest_head = odata + (iter*CREG_WIDTH)*ldo + (tid*num_cols_g + col);
						// load 8 real floating point nubmers from address
						// mt into a float64 vector.
						_reg = _MM512_LOAD_PPC( (void const *)loop_head);
						//					_mm512_load_pd( (void const *)loop_head);

						// this intrinsic needs int32 indices, but a 512 register
						// contains 16 int32 integers, and there is only one intrinsic
						// to set all 16 positoins of the register. So _index contains
						// 16 int32 integers.
						_MM512_I32LOEXTSCATTER_PPC((void *) dest_head, _index, _reg,
								_MM_DOWNCONV_PC_NONE, 1, _MM_HINT_NONE);
					}
				}
			}	
		}
	} else {
		// init index[], it's for scattering the entries to the dest matrix
		for (i = 0; i < 16; i++) {
			index[i] = i*CREG_WIDTH*ldo;
		}

		// 62 cores, 4 hardware threads per core. Leave out one core for the OS
		// (62 - 1)*4 = 244
		num_threads = 244;
		// number of rows in each group
		num_rows_g = (m/CREG_WIDTH + num_threads - 1)/num_threads*CREG_WIDTH;
		_index = _mm512_set_epi32(index[15], index[14], index[13], index[12],
				index[11], index[10], index[9], index[8], index[7], index[6],
				index[5], index[4], index[3], index[2], index[1], index[0]);

#pragma omp parallel num_threads(num_threads) private(tid, col, num_rows, iter,\
		loop_head, dest_head, _reg, num_iter)
		{
			tid = omp_get_thread_num();
			// compute the number of rows I need to work on
			num_rows = num_rows_g;
			if (tid == (num_threads - 1) && (m%num_rows_g != 0)) {
				num_rows = m%num_rows_g;
			}
			num_iter = (int) num_rows*1.0/CREG_WIDTH;

			for (iter = 0; iter < num_iter; iter++) {
				if (tid*num_rows_g + iter*CREG_WIDTH < m) {
					for (col = 0; col < n; col++) {
						loop_head = idata + (tid*num_rows_g + iter*CREG_WIDTH) + ldi*col;
						dest_head = odata + ldo*(tid*num_rows_g + iter*CREG_WIDTH) + col;

						_reg = _MM512_LOAD_PPC((void const *) loop_head);
						_MM512_I32LOEXTSCATTER_PPC((void *) dest_head, _index, _reg,
								_MM_DOWNCONV_PC_NONE, 1, _MM_HINT_NONE);
					}
				}
			}

		}
	}
	return;
}

void call_ctranspose_mkl(magma_mic_ctranspose_param ctranspose_param)
{
	magmaFloatComplex *odata, *idata;
	int ldo,ldi, m, n;
	int num_threads;
	int tid;
	int index[16], i;
	//  _M512 _reg;
	//  _M512I _index;
	magmaFloatComplex *my_head, *loop_head, *dest_head;
	int num_cols, num_cols_g, col, iter, num_iter;

	odata =(magmaFloatComplex *) ctranspose_param.odata;
	ldo   =ctranspose_param.ldo;
	idata =(magmaFloatComplex *) ctranspose_param.idata;
	ldi   = ctranspose_param.ldi;
	m     = ctranspose_param.m;
	n     = ctranspose_param.n;
	magmaFloatComplex alpha = MAGMA_C_MAKE(  1.0, 0.0 );

	mkl_comatcopy('C', 'T', m, n, *((MKL_Complex8 *)(&alpha)), (MKL_Complex8 *)idata, ldi, 
			(MKL_Complex8 *)odata, ldo);

	return;
}

//void call_claswp(magma_mic_claswp_param claswp_param) {
	/*
	 * takes 0-based ipiv
	 * In magma semantics, d_AT is always the starting point of the matrix, it
	 * doesn't change.
	 */
	//	fprintf(stdout, "entering claswp\n");
//	float *d_AT;
//	int lda, nb, ind;
//	int *d_ipiv;

//	d_AT		=	(float *) claswp_param.d_AT;
//	lda			=	claswp_param.lda;
//	d_ipiv		=	(int *) claswp_param.d_ipiv;
//	nb			=	claswp_param.nb;
//	ind			=	claswp_param.ind;

//	if (nb == 0) {
//		return;
//	}

//	int my_i, my_j;
//	int tid;
//	float *p1, *p2, tmp, *head;
//	int  *ipiv_tmp;
//	int num_threads;
//	int blk, num_blk;

//	__M512C _reg1, _reg2;

//	ipiv_tmp = d_ipiv;
	// (62 - 1)*4 = 244
//	num_threads		= 244;
	// the number of 8-column blocks in each section. Every thread takes one
	// section.
//	num_blk			= (lda/CREG_WIDTH + num_threads - 1)/num_threads;
	//	printf("num_blk = %d\n", num_blk);
//#pragma omp parallel num_threads(num_threads) private(tid, p1, p2, tmp, head,\
//		my_i, my_j, blk, _reg1, _reg2)
//	{
		/* 
		 * num_reps		the number of 512 vectors this thread owns
		 */
//		tid = omp_get_thread_num();
//		for (my_i = 0; my_i < nb; my_i++) { // only swap nb rows
			// my_j is the destination row index
			// convert 1-based to 0 based ipiv
//			my_j = d_ipiv[my_i + ind] - 1 + ind; 
//			if (my_i + ind == my_j) continue;
			// in my section, work on the 8-column blocks one by one
//			for (blk = 0; blk < num_blk; blk++) { // loop in a row
//				head = d_AT + tid*(num_blk*CREG_WIDTH) + blk*CREG_WIDTH;
				// check the end of this thread is in range
//				if (tid*(num_blk*CREG_WIDTH) + blk*CREG_WIDTH < lda) { 
					// p1 -- the source element
					// p2 -- the dest element
//					p1 = head + (my_i + ind)*lda;
//					p2 = head + my_j*lda;
//					_reg1 = _MM512_LOAD_PPC( (void const *) p1);
//					_reg2 = _MM512_LOAD_PPC( (void const *) p2);
//					_MM512_STORE_PPC( (void *) p2, _reg1);
//					_MM512_STORE_PPC( (void *) p1, _reg2);
//				}
//			}
//		}
//	}
//	return;
//}

void call_claswp(magma_mic_claswp_param claswp_param) {
	magma_int_t n, k1, k2, lda, incx;
	magmaFloatComplex *d_A;
	magma_int_t *d_ipiv;

	n		=	claswp_param.n;
	d_A		=	(magmaFloatComplex *) claswp_param.d_A;
	lda		=	claswp_param.lda;
	k1		=	claswp_param.k1;
	k2		=	claswp_param.k2;
	d_ipiv		=	(magma_int_t *) claswp_param.d_ipiv;
	incx		=	claswp_param.incx;

	//printf("n: %d lda: %d k1: %d k2: %d incx: %d\n", n, lda, k1, k2, incx);
	claswp(&n, (MKL_Complex8 *)d_A, &lda, &k1, &k2, (magma_int_t *)d_ipiv, &incx);
	return;
}
