/*
 *   -- micMAGMA (version 0.1) --
 *      Univ. of Tennessee, Knoxville
 *      Univ. of California, Berkeley
 *      Univ. of Colorado, Denver
 *      April 2012
 *
 * @generated d Thu Dec 13 14:39:22 2012
 */

#include <mkl.h>

#include <mkl_trans.h>
#include <scif.h>
#include <fcntl.h>
#include <offload.h>

//#ifdef __MIC__
//#define _M512_TYPED
//#include <lmmintrin.h>
//#endif
#include <immintrin.h>


#include "magma.h"
//#include "magma_d_bindings.h"
#include "magma_d_server.h"



void call_dgemm(magma_mic_dgemm_param dgemm_param) {
	magma_trans_t transa, transb;
	magma_int_t m, n, k, lda, ldb, ldc;
	double alpha, beta, *a, *b, *c;
	
        transa = dgemm_param.transa;
        transb = dgemm_param.transb;
        m = dgemm_param.m;
        n = dgemm_param.n;
        k = dgemm_param.k;
        alpha = dgemm_param.alpha;
        a = (double *)dgemm_param.a;
        lda = dgemm_param.lda;
        b = (double *)dgemm_param.b;
        ldb = dgemm_param.ldb;
        beta = dgemm_param.beta;
        c = (double *)dgemm_param.c;
        ldc = dgemm_param.ldc;

        //printf("DGEMM: %c %c %d %d %d %f 0x%lx %d 0x%lx %d %f 0x%lx %d\n", transa, transb, m, n, k, alpha, (off_t)a, lda, (off_t)b, ldb, beta, (off_t)c, ldc);

	dgemm( lapack_const(transa), lapack_const(transb), 
	       &m, &n, &k, (const double *)&alpha, (double *)a, &lda, 
	       (double *)b, &ldb, (const double *)&beta, 
	       (double *)c, &ldc);

	return;
}

void call_dgemv(magma_mic_dgemv_param dgemv_param) {
  magma_trans_t transa;
  magma_int_t m, n, lda, incx, incy;
  double alpha, beta, *a, *x, *y;

  transa = dgemv_param.transa;
  m      = dgemv_param.m;
  n      = dgemv_param.n;
  alpha  = dgemv_param.alpha;
  a      = (double *)dgemv_param.a;
  lda    = dgemv_param.lda;
  x      = (double *)dgemv_param.x;
  incx   = dgemv_param.incx;
  beta   = dgemv_param.beta;
  y      = (double *)dgemv_param.y;
  incy   = dgemv_param.incy;

  //printf("DGEMV: %c %c %d %d %d %f 0x%lx %d 0x%lx %d %f 0x%lx %d\n", transa, transb, m, n, k, alpha, (off_t)a, lda, (off_t)b, ldb, beta, (off_t)c, ldc);                                                                           

  dgemv( lapack_const(transa), &m, &n, (const double *)&alpha, 
	 (double *)a, &lda, (double *)x, &incx, 
	 (const double *)&beta, (double *)y, &incy);

  return;
}

void call_dtrsm(magma_mic_dtrsm_param dtrsm_param) {
        magma_side_t side;
	magma_uplo_t uplo;
	magma_trans_t transa;
	magma_diag_t diag;
	magma_int_t m, n, lda, ldb;
	double alpha, *a, *b;
	
        side = dtrsm_param.side;
        uplo = dtrsm_param.uplo;
        transa = dtrsm_param.transa;
        diag = dtrsm_param.diag;
        m = dtrsm_param.m;
        n = dtrsm_param.n;
        alpha = dtrsm_param.alpha;
        a = (double *)dtrsm_param.a;
        lda = dtrsm_param.lda;
        b = (double *)dtrsm_param.b;
        ldb = dtrsm_param.ldb;

        //printf("DTRSM: %c %c %c %c %d %d %f 0x%lx %d 0x%lx %d\n", side, uplo, transa, diag, m, n, alpha, (off_t)a, lda, (off_t)b, ldb);
	dtrsm( lapack_const(side), lapack_const(uplo), lapack_const(transa), lapack_const(diag),
	       &m, &n, (const double *)&alpha, (double *)a, &lda, 
	       (double *)b, &ldb);

	return;
}


void call_dtrmm(magma_mic_dtrmm_param dtrmm_param) {
        magma_side_t side;
	magma_uplo_t uplo;
	magma_trans_t transa;
	magma_diag_t diag;
        magma_int_t m, n, lda, ldb;
        double alpha, *a, *b;

        side = dtrmm_param.side;
        uplo = dtrmm_param.uplo;
        transa = dtrmm_param.transa;
        diag = dtrmm_param.diag;
        m = dtrmm_param.m;
        n = dtrmm_param.n;
        alpha = dtrmm_param.alpha;
        a = (double *)dtrmm_param.a;
        lda = dtrmm_param.lda;
        b = (double *)dtrmm_param.b;
        ldb = dtrmm_param.ldb;

        //printf("DTRMM: %c %c %c %c %d %d %f 0x%lx %d 0x%lx %d\n", side, uplo, transa, diag, m, n, alpha, (off_t)a, lda, (off_t)b, ldb);
	dtrmm(lapack_const(side), lapack_const(uplo), lapack_const(transa), lapack_const(diag),
	      &m, &n, (const double *)&alpha, (double *)a, &lda, 
	      (double *)b, &ldb);

	return;
}
        


void call_dsyrk(magma_mic_dsyrk_param dsyrk_param) {
        magma_uplo_t uplo;
	magma_trans_t trans;
	magma_int_t n, k, lda, ldc;
	double alpha, beta;
	double *a, *c;
	
        uplo = dsyrk_param.uplo;
        trans = dsyrk_param.trans;
        n = dsyrk_param.n;
        k = dsyrk_param.k;
        alpha = dsyrk_param.alpha;
        a = (double *)dsyrk_param.a;
        lda = dsyrk_param.lda;
        beta = dsyrk_param.beta;
        c = (double *)dsyrk_param.c;
        ldc = dsyrk_param.ldc;

        //printf("DSYRK: %c %c %d %d %f 0x%lx %d %f 0x%lx %d\n", uplo, trans, n, k, alpha, (off_t)a, lda, beta, (off_t)c, ldc);

	dsyrk(lapack_const(uplo), lapack_const(trans), 
	      &n, &k, &alpha, (double *)a, &lda, 
	      &beta, (double *)c, &ldc);

	return;
}


void call_dlarft(magma_mic_dlarft_param dlarft_param) {
	int direct, storev;
	magma_int_t n, k, ldv, ldt; 
	double *v, *tau, *t;
 
        direct = dlarft_param.direct;
        storev = dlarft_param.storev;
        n = dlarft_param.n;
        k = dlarft_param.k;
        v = (double *)dlarft_param.v;
        ldv = dlarft_param.ldv;
        tau = (double *)dlarft_param.tau;
        t = (double *)dlarft_param.t;
        ldt = dlarft_param.ldt;

	//printf("ZLARFT: %c %c %d %d 0x%lx %d ox%lx 0x%lx %d\n", direct, storev, n, k, (off_t)v, ldv, (off_t)tau, (off_t)t, ldt);

	dlarft(lapack_const(direct), lapack_const(storev),
	       &n, &k, (double *)v, &ldv, 
	       (double *)tau, (double *)t, &ldt);
	
	return;
}


void call_dlarfb(magma_mic_dlarfb_param dlarfb_param) {
	int side, trans, direct, storev;
	magma_int_t m, n, k, ldv, ldt, ldc, ldwork;
	double *v, *t, *c, *work;

        side = dlarfb_param.side;
        trans = dlarfb_param.trans;
        direct = dlarfb_param.direct;
        storev = dlarfb_param.storev;
        m = dlarfb_param.m;
        n = dlarfb_param.n;
        k = dlarfb_param.k;
        v = (double *)dlarfb_param.v;
        ldv = dlarfb_param.ldv;
        t = (double *)dlarfb_param.t;
        ldt = dlarfb_param.ldt;
        c = (double *)dlarfb_param.c;
        ldc = dlarfb_param.ldc;
        work = (double *)dlarfb_param.work;
        ldwork = dlarfb_param.ldwork;

        //printf("DLARFB: %c %c %c %c %d %d %d 0x%lx %d ox%lx %d 0x%lx %d 0x%lx %d\n", side, trans, direct, storev, m, n, k, (off_t)v, ldv, (off_t)t, ldt, (off_t)c, ldc, (off_t)work, ldwork); 
	dlarfb(lapack_const(side), lapack_const(trans), lapack_const(direct), lapack_const(storev),
			&m, &n, &k, (double *)v, &ldv, (double *)t, &ldt, 
			(double *)c, &ldc, (double *)work, &ldwork);

	/*
	double c_zero    = MAGMA_D_MAKE(  0.0, 0.0 );
	double c_one     = MAGMA_D_MAKE(  1.0, 0.0 );
	double c_neg_one = MAGMA_D_MAKE( -1.0, 0.0 );

	if (m <= 0 || n <= 0) {
		return;
	}

	magma_int_t transt;
	if (trans == MagmaNoTrans) 
		transt = MagmaTrans;
	else
		transt = MagmaNoTrans;

	if ( side  == MagmaLeft ) {

		if ( storev == MagmaColumnwise ) 
		{
			dgemm( lapack_const(MagmaTrans), lapack_const(MagmaNoTrans),
					&n, &k, &m,
					&c_one,  (double *)c,    &ldc,
					(double *)v,    &ldv,
					&c_zero, (double *)work, &ldwork);

			if (direct == MagmaForward)
				dtrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
						&n, &k, 
						&c_one, (double *)t,    &ldt, 
						(double *)work, &ldwork);
			else
				dtrmm( lapack_const(MagmaRight), lapack_const(MagmaLower), lapack_const(transt), lapack_const(MagmaNonUnit),
						&n, &k, 
						&c_one, (double *)t,    &ldt, 
						(double *)work, &ldwork);

			dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), 
					&m, &n, &k, 
					&c_neg_one, (double *)v,    &ldv,
					(double *)work, &ldwork, 
					&c_one,     (double *)c,    &ldc);
		}
		else {
			dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), 
					&m, &k, &n, 
					&c_one,  (double *)c,    &ldc,
					(double *)v,    &ldv, 
					&c_zero, (double *)work, &ldwork);

			dtrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
					&m, &k, 
					&c_one, (double *)t,    &ldt, 
					(double *)work, &ldwork);

			dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans), 
					&m, &n, &k, 
					&c_neg_one, (double *)work, &ldwork,
					(double *)v,    &ldv,
					&c_one,     (double *)c,    &ldc);
		}
	}

	else {

		if ( storev == MagmaColumnwise ) {
			dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans),
					&m, &k, &n,
					&c_one,  (double *)c,    &ldc,
					(double *)v,    &ldv,
					&c_zero, (double *)work, &ldwork);
			// ??? ldwork replaced by k for case n < k

			if (direct == MagmaForward)
				dtrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
						&m, &k,
						&c_one, (double *)t,    &ldt,
						(double *)work, &ldwork);
			else
				dtrmm( lapack_const(MagmaRight), lapack_const(MagmaLower), lapack_const(transt), lapack_const(MagmaNonUnit),
						&m, &k,
						&c_one, (double *)t,    &ldt,
						(double *)work, &ldwork);

			dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans),
					&m, &n, &k,
					&c_neg_one, (double *)work, &ldwork, 
					(double *)v,    &ldv,
					&c_one,     (double *)c,    &ldc);
		}
		else {
			dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans),
					&m, &k, &n,
					&c_one,  (double *)c,    &ldc,
					(double *)v,    &ldv,
					&c_zero, (double *)work, &ldwork);

			dtrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
					&m, &k,
					&c_one, (double *)t,    &ldt,
					(double *)work, &ldwork);

			dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans),
					&m, &n, &k,
					&c_neg_one, (double *)work, &ldwork,
					(double *)v,    &ldv,
					&c_one,     (double *)c,    &ldc);
		}
	}*/

	return;
}


void call_dtranspose(magma_mic_dtranspose_param dtranspose_param)
{
	/******************************************************
	 * _mm512_load_pd -- for loading
	 * _mm512_i32lo[ext]scatter_pd
	 * _mm512_i32loscatter_pd is the simplified version of ...ext...
	 * they do the same thing, the latter needs less arguments.
	 * **************************************************/
	//	printf("entered stranpose\n");
	//	fflush(stdout);
	double *odata, *idata;
	int ldo,ldi, m, n;
	int num_threads;
	int tid;
	int index[16], i;
	__M512D _reg;
	__m512i _index;
	double *my_head, *loop_head, *dest_head;
	int num_cols, num_cols_g, col, iter, num_iter, num_rows_g, num_rows;

	odata	=	(double *) dtranspose_param.odata;
	ldo		=	dtranspose_param.ldo;
	idata	=	(double *) dtranspose_param.idata;
	ldi		=	dtranspose_param.ldi;
	m		=	dtranspose_param.m;
	n		=	dtranspose_param.n;

	if (m == 0 || n == 0) {
		return;
	}
	if (m <= n) {
		for (i = 0; i < 16; i++) {
			index[i] = i*DREG_WIDTH*ldo;
		}
		// 62 cores, 4 hardware threads per core. Leave out one core for the OS
		// (62 - 1)*4 = 244

		num_threads = 244;
		num_cols_g =	(n + num_threads - 1)/num_threads; 
		num_iter = (int) m*1.0/DREG_WIDTH;
		_index = _mm512_set_epi32(index[15], index[14], index[13], index[12],
				index[11], index[10], index[9], index[8], index[7], index[6],
				index[5], index[4], index[3], index[2], index[1], index[0]);

#pragma omp parallel num_threads(num_threads) private(tid, col, num_cols, iter,\
		my_head, loop_head, dest_head, _reg)
		{
			tid = omp_get_thread_num();
			num_cols = num_cols_g;
			if (tid == (num_threads - 1) && (n%num_cols_g != 0)) { 
				num_cols = n%num_cols_g;
			}
			// starting point of my section, constant for each thread
			my_head = idata + ldi*(tid*num_cols_g);
			for (iter = 0; iter < num_iter; iter++) { // loop inside each column
				for (col = 0; col < num_cols; col++) { // loop over each column
					if (tid*num_cols_g + col < n) {
						// starting point of the inner loop
						loop_head = my_head + col*ldi + iter*DREG_WIDTH;
						dest_head = odata + (iter*DREG_WIDTH)*ldo + (tid*num_cols_g + col);
						// load 8 double precision floating point nubmers from address
						// mt into a float64 vector.
						_reg = _MM512_LOAD_PPD( (void const *)loop_head);
						//					_mm512_load_pd( (void const *)loop_head);

						// this intrinsic needs int32 indices, but a 512 register
						// contains 16 int32 integers, and there is only one intrinsic
						// to set all 16 positoins of the register. So _index contains
						// 16 int32 integers.
						_MM512_I32LOEXTSCATTER_PPD((void *) dest_head, _index, _reg,
								_MM_DOWNCONV_PD_NONE, 1, _MM_HINT_NONE);
					}
				}
			}	
		}
	} else {
		// init index[], it's for scattering the entries to the dest matrix
		for (i = 0; i < 16; i++) {
			index[i] = i*DREG_WIDTH*ldo;
		}

		// 62 cores, 4 hardware threads per core. Leave out one core for the OS
		// (62 - 1)*4 = 244
		num_threads = 244;
		// number of rows in each group
		num_rows_g = (m/DREG_WIDTH + num_threads - 1)/num_threads*DREG_WIDTH;
		_index = _mm512_set_epi32(index[15], index[14], index[13], index[12],
				index[11], index[10], index[9], index[8], index[7], index[6],
				index[5], index[4], index[3], index[2], index[1], index[0]);

#pragma omp parallel num_threads(num_threads) private(tid, col, num_rows, iter,\
		loop_head, dest_head, _reg, num_iter)
		{
			tid = omp_get_thread_num();
			// compute the number of rows I need to work on
			num_rows = num_rows_g;
			if (tid == (num_threads - 1) && (m%num_rows_g != 0)) {
				num_rows = m%num_rows_g;
			}
			num_iter = (int) num_rows*1.0/DREG_WIDTH;

			for (iter = 0; iter < num_iter; iter++) {
				if (tid*num_rows_g + iter*DREG_WIDTH < m) {
					for (col = 0; col < n; col++) {
						loop_head = idata + (tid*num_rows_g + iter*DREG_WIDTH) + ldi*col;
						dest_head = odata + ldo*(tid*num_rows_g + iter*DREG_WIDTH) + col;

						_reg = _MM512_LOAD_PPD((void const *) loop_head);
						_MM512_I32LOEXTSCATTER_PPD((void *) dest_head, _index, _reg,
								_MM_DOWNCONV_PD_NONE, 1, _MM_HINT_NONE);
					}
				}
			}

		}
	}
	return;
}

void call_dtranspose_mkl(magma_mic_dtranspose_param dtranspose_param)
{
	double *odata, *idata;
	int ldo,ldi, m, n;
	int num_threads;
	int tid;
	int index[16], i;
	//  _M512 _reg;
	//  _M512I _index;
	double *my_head, *loop_head, *dest_head;
	int num_cols, num_cols_g, col, iter, num_iter;

	odata =(double *) dtranspose_param.odata;
	ldo   =dtranspose_param.ldo;
	idata =(double *) dtranspose_param.idata;
	ldi   = dtranspose_param.ldi;
	m     = dtranspose_param.m;
	n     = dtranspose_param.n;
	double alpha = MAGMA_D_MAKE(  1.0, 0.0 );

	mkl_domatcopy('C', 'T', m, n, *((double *)(&alpha)), (double *)idata, ldi, 
			(double *)odata, ldo);

	return;
}

//void call_dlaswp(magma_mic_dlaswp_param dlaswp_param) {
	/*
	 * takes 0-based ipiv
	 * In magma semantics, d_AT is always the starting point of the matrix, it
	 * doesn't change.
	 */
	//	fprintf(stdout, "entering dlaswp\n");
//	double *d_AT;
//	int lda, nb, ind;
//	int *d_ipiv;

//	d_AT		=	(double *) dlaswp_param.d_AT;
//	lda			=	dlaswp_param.lda;
//	d_ipiv		=	(int *) dlaswp_param.d_ipiv;
//	nb			=	dlaswp_param.nb;
//	ind			=	dlaswp_param.ind;

//	if (nb == 0) {
//		return;
//	}

//	int my_i, my_j;
//	int tid;
//	double *p1, *p2, tmp, *head;
//	int  *ipiv_tmp;
//	int num_threads;
//	int blk, num_blk;

//	__M512D _reg1, _reg2;

//	ipiv_tmp = d_ipiv;
	// (62 - 1)*4 = 244
//	num_threads		= 244;
	// the number of 8-column blocks in each section. Every thread takes one
	// section.
//	num_blk			= (lda/DREG_WIDTH + num_threads - 1)/num_threads;
	//	printf("num_blk = %d\n", num_blk);
//#pragma omp parallel num_threads(num_threads) private(tid, p1, p2, tmp, head,\
//		my_i, my_j, blk, _reg1, _reg2)
//	{
		/* 
		 * num_reps		the number of 512 vectors this thread owns
		 */
//		tid = omp_get_thread_num();
//		for (my_i = 0; my_i < nb; my_i++) { // only swap nb rows
			// my_j is the destination row index
			// convert 1-based to 0 based ipiv
//			my_j = d_ipiv[my_i + ind] - 1 + ind; 
//			if (my_i + ind == my_j) continue;
			// in my section, work on the 8-column blocks one by one
//			for (blk = 0; blk < num_blk; blk++) { // loop in a row
//				head = d_AT + tid*(num_blk*DREG_WIDTH) + blk*DREG_WIDTH;
				// check the end of this thread is in range
//				if (tid*(num_blk*DREG_WIDTH) + blk*DREG_WIDTH < lda) { 
					// p1 -- the source element
					// p2 -- the dest element
//					p1 = head + (my_i + ind)*lda;
//					p2 = head + my_j*lda;
//					_reg1 = _MM512_LOAD_PPD( (void const *) p1);
//					_reg2 = _MM512_LOAD_PPD( (void const *) p2);
//					_MM512_STORE_PPD( (void *) p2, _reg1);
//					_MM512_STORE_PPD( (void *) p1, _reg2);
//				}
//			}
//		}
//	}
//	return;
//}

void call_dlaswp(magma_mic_dlaswp_param dlaswp_param) {
	magma_int_t n, k1, k2, lda, incx;
	double *d_A;
	magma_int_t *d_ipiv;

	n		=	dlaswp_param.n;
	d_A		=	(double *) dlaswp_param.d_A;
	lda		=	dlaswp_param.lda;
	k1		=	dlaswp_param.k1;
	k2		=	dlaswp_param.k2;
	d_ipiv		=	(magma_int_t *) dlaswp_param.d_ipiv;
	incx		=	dlaswp_param.incx;

	//printf("n: %d lda: %d k1: %d k2: %d incx: %d\n", n, lda, k1, k2, incx);
	dlaswp(&n, (double *)d_A, &lda, &k1, &k2, (magma_int_t *)d_ipiv, &incx);
	return;
}
