/*
 *   -- micMAGMA (version 0.1) --
 *      Univ. of Tennessee, Knoxville
 *      Univ. of California, Berkeley
 *      Univ. of Colorado, Denver
 *      April 2012
 *
 * @generated s Thu Dec 13 14:39:22 2012
 */

#include <mkl.h>

#include <mkl_trans.h>
#include <scif.h>
#include <fcntl.h>
#include <offload.h>

//#ifdef __MIC__
//#define _M512_TYPED
//#include <lmmintrin.h>
//#endif
#include <immintrin.h>


#include "magma.h"
//#include "magma_s_bindings.h"
#include "magma_s_server.h"



void call_sgemm(magma_mic_sgemm_param sgemm_param) {
	magma_trans_t transa, transb;
	magma_int_t m, n, k, lda, ldb, ldc;
	float alpha, beta, *a, *b, *c;
	
        transa = sgemm_param.transa;
        transb = sgemm_param.transb;
        m = sgemm_param.m;
        n = sgemm_param.n;
        k = sgemm_param.k;
        alpha = sgemm_param.alpha;
        a = (float *)sgemm_param.a;
        lda = sgemm_param.lda;
        b = (float *)sgemm_param.b;
        ldb = sgemm_param.ldb;
        beta = sgemm_param.beta;
        c = (float *)sgemm_param.c;
        ldc = sgemm_param.ldc;

        //printf("SGEMM: %c %c %d %d %d %f 0x%lx %d 0x%lx %d %f 0x%lx %d\n", transa, transb, m, n, k, alpha, (off_t)a, lda, (off_t)b, ldb, beta, (off_t)c, ldc);

	sgemm( lapack_const(transa), lapack_const(transb), 
	       &m, &n, &k, (const float *)&alpha, (float *)a, &lda, 
	       (float *)b, &ldb, (const float *)&beta, 
	       (float *)c, &ldc);

	return;
}

void call_sgemv(magma_mic_sgemv_param sgemv_param) {
  magma_trans_t transa;
  magma_int_t m, n, lda, incx, incy;
  float alpha, beta, *a, *x, *y;

  transa = sgemv_param.transa;
  m      = sgemv_param.m;
  n      = sgemv_param.n;
  alpha  = sgemv_param.alpha;
  a      = (float *)sgemv_param.a;
  lda    = sgemv_param.lda;
  x      = (float *)sgemv_param.x;
  incx   = sgemv_param.incx;
  beta   = sgemv_param.beta;
  y      = (float *)sgemv_param.y;
  incy   = sgemv_param.incy;

  //printf("SGEMV: %c %c %d %d %d %f 0x%lx %d 0x%lx %d %f 0x%lx %d\n", transa, transb, m, n, k, alpha, (off_t)a, lda, (off_t)b, ldb, beta, (off_t)c, ldc);                                                                           

  sgemv( lapack_const(transa), &m, &n, (const float *)&alpha, 
	 (float *)a, &lda, (float *)x, &incx, 
	 (const float *)&beta, (float *)y, &incy);

  return;
}

void call_strsm(magma_mic_strsm_param strsm_param) {
        magma_side_t side;
	magma_uplo_t uplo;
	magma_trans_t transa;
	magma_diag_t diag;
	magma_int_t m, n, lda, ldb;
	float alpha, *a, *b;
	
        side = strsm_param.side;
        uplo = strsm_param.uplo;
        transa = strsm_param.transa;
        diag = strsm_param.diag;
        m = strsm_param.m;
        n = strsm_param.n;
        alpha = strsm_param.alpha;
        a = (float *)strsm_param.a;
        lda = strsm_param.lda;
        b = (float *)strsm_param.b;
        ldb = strsm_param.ldb;

        //printf("STRSM: %c %c %c %c %d %d %f 0x%lx %d 0x%lx %d\n", side, uplo, transa, diag, m, n, alpha, (off_t)a, lda, (off_t)b, ldb);
	strsm( lapack_const(side), lapack_const(uplo), lapack_const(transa), lapack_const(diag),
	       &m, &n, (const float *)&alpha, (float *)a, &lda, 
	       (float *)b, &ldb);

	return;
}


void call_strmm(magma_mic_strmm_param strmm_param) {
        magma_side_t side;
	magma_uplo_t uplo;
	magma_trans_t transa;
	magma_diag_t diag;
        magma_int_t m, n, lda, ldb;
        float alpha, *a, *b;

        side = strmm_param.side;
        uplo = strmm_param.uplo;
        transa = strmm_param.transa;
        diag = strmm_param.diag;
        m = strmm_param.m;
        n = strmm_param.n;
        alpha = strmm_param.alpha;
        a = (float *)strmm_param.a;
        lda = strmm_param.lda;
        b = (float *)strmm_param.b;
        ldb = strmm_param.ldb;

        //printf("STRMM: %c %c %c %c %d %d %f 0x%lx %d 0x%lx %d\n", side, uplo, transa, diag, m, n, alpha, (off_t)a, lda, (off_t)b, ldb);
	strmm(lapack_const(side), lapack_const(uplo), lapack_const(transa), lapack_const(diag),
	      &m, &n, (const float *)&alpha, (float *)a, &lda, 
	      (float *)b, &ldb);

	return;
}
        


void call_ssyrk(magma_mic_ssyrk_param ssyrk_param) {
        magma_uplo_t uplo;
	magma_trans_t trans;
	magma_int_t n, k, lda, ldc;
	float alpha, beta;
	float *a, *c;
	
        uplo = ssyrk_param.uplo;
        trans = ssyrk_param.trans;
        n = ssyrk_param.n;
        k = ssyrk_param.k;
        alpha = ssyrk_param.alpha;
        a = (float *)ssyrk_param.a;
        lda = ssyrk_param.lda;
        beta = ssyrk_param.beta;
        c = (float *)ssyrk_param.c;
        ldc = ssyrk_param.ldc;

        //printf("SSYRK: %c %c %d %d %f 0x%lx %d %f 0x%lx %d\n", uplo, trans, n, k, alpha, (off_t)a, lda, beta, (off_t)c, ldc);

	ssyrk(lapack_const(uplo), lapack_const(trans), 
	      &n, &k, &alpha, (float *)a, &lda, 
	      &beta, (float *)c, &ldc);

	return;
}


void call_slarft(magma_mic_slarft_param slarft_param) {
	int direct, storev;
	magma_int_t n, k, ldv, ldt; 
	float *v, *tau, *t;
 
        direct = slarft_param.direct;
        storev = slarft_param.storev;
        n = slarft_param.n;
        k = slarft_param.k;
        v = (float *)slarft_param.v;
        ldv = slarft_param.ldv;
        tau = (float *)slarft_param.tau;
        t = (float *)slarft_param.t;
        ldt = slarft_param.ldt;

	//printf("ZLARFT: %c %c %d %d 0x%lx %d ox%lx 0x%lx %d\n", direct, storev, n, k, (off_t)v, ldv, (off_t)tau, (off_t)t, ldt);

	slarft(lapack_const(direct), lapack_const(storev),
	       &n, &k, (float *)v, &ldv, 
	       (float *)tau, (float *)t, &ldt);
	
	return;
}


void call_slarfb(magma_mic_slarfb_param slarfb_param) {
	int side, trans, direct, storev;
	magma_int_t m, n, k, ldv, ldt, ldc, ldwork;
	float *v, *t, *c, *work;

        side = slarfb_param.side;
        trans = slarfb_param.trans;
        direct = slarfb_param.direct;
        storev = slarfb_param.storev;
        m = slarfb_param.m;
        n = slarfb_param.n;
        k = slarfb_param.k;
        v = (float *)slarfb_param.v;
        ldv = slarfb_param.ldv;
        t = (float *)slarfb_param.t;
        ldt = slarfb_param.ldt;
        c = (float *)slarfb_param.c;
        ldc = slarfb_param.ldc;
        work = (float *)slarfb_param.work;
        ldwork = slarfb_param.ldwork;

        //printf("SLARFB: %c %c %c %c %d %d %d 0x%lx %d ox%lx %d 0x%lx %d 0x%lx %d\n", side, trans, direct, storev, m, n, k, (off_t)v, ldv, (off_t)t, ldt, (off_t)c, ldc, (off_t)work, ldwork); 
	slarfb(lapack_const(side), lapack_const(trans), lapack_const(direct), lapack_const(storev),
			&m, &n, &k, (float *)v, &ldv, (float *)t, &ldt, 
			(float *)c, &ldc, (float *)work, &ldwork);

	/*
	float c_zero    = MAGMA_S_MAKE(  0.0, 0.0 );
	float c_one     = MAGMA_S_MAKE(  1.0, 0.0 );
	float c_neg_one = MAGMA_S_MAKE( -1.0, 0.0 );

	if (m <= 0 || n <= 0) {
		return;
	}

	magma_int_t transt;
	if (trans == MagmaNoTrans) 
		transt = MagmaTrans;
	else
		transt = MagmaNoTrans;

	if ( side  == MagmaLeft ) {

		if ( storev == MagmaColumnwise ) 
		{
			sgemm( lapack_const(MagmaTrans), lapack_const(MagmaNoTrans),
					&n, &k, &m,
					&c_one,  (float *)c,    &ldc,
					(float *)v,    &ldv,
					&c_zero, (float *)work, &ldwork);

			if (direct == MagmaForward)
				strmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
						&n, &k, 
						&c_one, (float *)t,    &ldt, 
						(float *)work, &ldwork);
			else
				strmm( lapack_const(MagmaRight), lapack_const(MagmaLower), lapack_const(transt), lapack_const(MagmaNonUnit),
						&n, &k, 
						&c_one, (float *)t,    &ldt, 
						(float *)work, &ldwork);

			sgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), 
					&m, &n, &k, 
					&c_neg_one, (float *)v,    &ldv,
					(float *)work, &ldwork, 
					&c_one,     (float *)c,    &ldc);
		}
		else {
			sgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), 
					&m, &k, &n, 
					&c_one,  (float *)c,    &ldc,
					(float *)v,    &ldv, 
					&c_zero, (float *)work, &ldwork);

			strmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
					&m, &k, 
					&c_one, (float *)t,    &ldt, 
					(float *)work, &ldwork);

			sgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans), 
					&m, &n, &k, 
					&c_neg_one, (float *)work, &ldwork,
					(float *)v,    &ldv,
					&c_one,     (float *)c,    &ldc);
		}
	}

	else {

		if ( storev == MagmaColumnwise ) {
			sgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans),
					&m, &k, &n,
					&c_one,  (float *)c,    &ldc,
					(float *)v,    &ldv,
					&c_zero, (float *)work, &ldwork);
			// ??? ldwork replaced by k for case n < k

			if (direct == MagmaForward)
				strmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
						&m, &k,
						&c_one, (float *)t,    &ldt,
						(float *)work, &ldwork);
			else
				strmm( lapack_const(MagmaRight), lapack_const(MagmaLower), lapack_const(transt), lapack_const(MagmaNonUnit),
						&m, &k,
						&c_one, (float *)t,    &ldt,
						(float *)work, &ldwork);

			sgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans),
					&m, &n, &k,
					&c_neg_one, (float *)work, &ldwork, 
					(float *)v,    &ldv,
					&c_one,     (float *)c,    &ldc);
		}
		else {
			sgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans),
					&m, &k, &n,
					&c_one,  (float *)c,    &ldc,
					(float *)v,    &ldv,
					&c_zero, (float *)work, &ldwork);

			strmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
					&m, &k,
					&c_one, (float *)t,    &ldt,
					(float *)work, &ldwork);

			sgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans),
					&m, &n, &k,
					&c_neg_one, (float *)work, &ldwork,
					(float *)v,    &ldv,
					&c_one,     (float *)c,    &ldc);
		}
	}*/

	return;
}


void call_stranspose(magma_mic_stranspose_param stranspose_param)
{
	/******************************************************
	 * _mm512_load_pd -- for loading
	 * _mm512_i32lo[ext]scatter_pd
	 * _mm512_i32loscatter_pd is the simplified version of ...ext...
	 * they do the same thing, the latter needs less arguments.
	 * **************************************************/
	//	printf("entered stranpose\n");
	//	fflush(stdout);
	float *odata, *idata;
	int ldo,ldi, m, n;
	int num_threads;
	int tid;
	int index[16], i;
	__M512S _reg;
	__m512i _index;
	float *my_head, *loop_head, *dest_head;
	int num_cols, num_cols_g, col, iter, num_iter, num_rows_g, num_rows;

	odata	=	(float *) stranspose_param.odata;
	ldo		=	stranspose_param.ldo;
	idata	=	(float *) stranspose_param.idata;
	ldi		=	stranspose_param.ldi;
	m		=	stranspose_param.m;
	n		=	stranspose_param.n;

	if (m == 0 || n == 0) {
		return;
	}
	if (m <= n) {
		for (i = 0; i < 16; i++) {
			index[i] = i*SREG_WIDTH*ldo;
		}
		// 62 cores, 4 hardware threads per core. Leave out one core for the OS
		// (62 - 1)*4 = 244

		num_threads = 244;
		num_cols_g =	(n + num_threads - 1)/num_threads; 
		num_iter = (int) m*1.0/SREG_WIDTH;
		_index = _mm512_set_epi32(index[15], index[14], index[13], index[12],
				index[11], index[10], index[9], index[8], index[7], index[6],
				index[5], index[4], index[3], index[2], index[1], index[0]);

#pragma omp parallel num_threads(num_threads) private(tid, col, num_cols, iter,\
		my_head, loop_head, dest_head, _reg)
		{
			tid = omp_get_thread_num();
			num_cols = num_cols_g;
			if (tid == (num_threads - 1) && (n%num_cols_g != 0)) { 
				num_cols = n%num_cols_g;
			}
			// starting point of my section, constant for each thread
			my_head = idata + ldi*(tid*num_cols_g);
			for (iter = 0; iter < num_iter; iter++) { // loop inside each column
				for (col = 0; col < num_cols; col++) { // loop over each column
					if (tid*num_cols_g + col < n) {
						// starting point of the inner loop
						loop_head = my_head + col*ldi + iter*SREG_WIDTH;
						dest_head = odata + (iter*SREG_WIDTH)*ldo + (tid*num_cols_g + col);
						// load 8 real floating point nubmers from address
						// mt into a float64 vector.
						_reg = _MM512_LOAD_PPS( (void const *)loop_head);
						//					_mm512_load_pd( (void const *)loop_head);

						// this intrinsic needs int32 indices, but a 512 register
						// contains 16 int32 integers, and there is only one intrinsic
						// to set all 16 positoins of the register. So _index contains
						// 16 int32 integers.
						_MM512_I32LOEXTSCATTER_PPS((void *) dest_head, _index, _reg,
								_MM_DOWNCONV_PS_NONE, 1, _MM_HINT_NONE);
					}
				}
			}	
		}
	} else {
		// init index[], it's for scattering the entries to the dest matrix
		for (i = 0; i < 16; i++) {
			index[i] = i*SREG_WIDTH*ldo;
		}

		// 62 cores, 4 hardware threads per core. Leave out one core for the OS
		// (62 - 1)*4 = 244
		num_threads = 244;
		// number of rows in each group
		num_rows_g = (m/SREG_WIDTH + num_threads - 1)/num_threads*SREG_WIDTH;
		_index = _mm512_set_epi32(index[15], index[14], index[13], index[12],
				index[11], index[10], index[9], index[8], index[7], index[6],
				index[5], index[4], index[3], index[2], index[1], index[0]);

#pragma omp parallel num_threads(num_threads) private(tid, col, num_rows, iter,\
		loop_head, dest_head, _reg, num_iter)
		{
			tid = omp_get_thread_num();
			// compute the number of rows I need to work on
			num_rows = num_rows_g;
			if (tid == (num_threads - 1) && (m%num_rows_g != 0)) {
				num_rows = m%num_rows_g;
			}
			num_iter = (int) num_rows*1.0/SREG_WIDTH;

			for (iter = 0; iter < num_iter; iter++) {
				if (tid*num_rows_g + iter*SREG_WIDTH < m) {
					for (col = 0; col < n; col++) {
						loop_head = idata + (tid*num_rows_g + iter*SREG_WIDTH) + ldi*col;
						dest_head = odata + ldo*(tid*num_rows_g + iter*SREG_WIDTH) + col;

						_reg = _MM512_LOAD_PPS((void const *) loop_head);
						_MM512_I32LOEXTSCATTER_PPS((void *) dest_head, _index, _reg,
								_MM_DOWNCONV_PS_NONE, 1, _MM_HINT_NONE);
					}
				}
			}

		}
	}
	return;
}

void call_stranspose_mkl(magma_mic_stranspose_param stranspose_param)
{
	float *odata, *idata;
	int ldo,ldi, m, n;
	int num_threads;
	int tid;
	int index[16], i;
	//  _M512 _reg;
	//  _M512I _index;
	float *my_head, *loop_head, *dest_head;
	int num_cols, num_cols_g, col, iter, num_iter;

	odata =(float *) stranspose_param.odata;
	ldo   =stranspose_param.ldo;
	idata =(float *) stranspose_param.idata;
	ldi   = stranspose_param.ldi;
	m     = stranspose_param.m;
	n     = stranspose_param.n;
	float alpha = MAGMA_S_MAKE(  1.0, 0.0 );

	mkl_somatcopy('C', 'T', m, n, *((float *)(&alpha)), (float *)idata, ldi, 
			(float *)odata, ldo);

	return;
}

//void call_slaswp(magma_mic_slaswp_param slaswp_param) {
	/*
	 * takes 0-based ipiv
	 * In magma semantics, d_AT is always the starting point of the matrix, it
	 * doesn't change.
	 */
	//	fprintf(stdout, "entering slaswp\n");
//	float *d_AT;
//	int lda, nb, ind;
//	int *d_ipiv;

//	d_AT		=	(float *) slaswp_param.d_AT;
//	lda			=	slaswp_param.lda;
//	d_ipiv		=	(int *) slaswp_param.d_ipiv;
//	nb			=	slaswp_param.nb;
//	ind			=	slaswp_param.ind;

//	if (nb == 0) {
//		return;
//	}

//	int my_i, my_j;
//	int tid;
//	float *p1, *p2, tmp, *head;
//	int  *ipiv_tmp;
//	int num_threads;
//	int blk, num_blk;

//	__M512S _reg1, _reg2;

//	ipiv_tmp = d_ipiv;
	// (62 - 1)*4 = 244
//	num_threads		= 244;
	// the number of 8-column blocks in each section. Every thread takes one
	// section.
//	num_blk			= (lda/SREG_WIDTH + num_threads - 1)/num_threads;
	//	printf("num_blk = %d\n", num_blk);
//#pragma omp parallel num_threads(num_threads) private(tid, p1, p2, tmp, head,\
//		my_i, my_j, blk, _reg1, _reg2)
//	{
		/* 
		 * num_reps		the number of 512 vectors this thread owns
		 */
//		tid = omp_get_thread_num();
//		for (my_i = 0; my_i < nb; my_i++) { // only swap nb rows
			// my_j is the destination row index
			// convert 1-based to 0 based ipiv
//			my_j = d_ipiv[my_i + ind] - 1 + ind; 
//			if (my_i + ind == my_j) continue;
			// in my section, work on the 8-column blocks one by one
//			for (blk = 0; blk < num_blk; blk++) { // loop in a row
//				head = d_AT + tid*(num_blk*SREG_WIDTH) + blk*SREG_WIDTH;
				// check the end of this thread is in range
//				if (tid*(num_blk*SREG_WIDTH) + blk*SREG_WIDTH < lda) { 
					// p1 -- the source element
					// p2 -- the dest element
//					p1 = head + (my_i + ind)*lda;
//					p2 = head + my_j*lda;
//					_reg1 = _MM512_LOAD_PPS( (void const *) p1);
//					_reg2 = _MM512_LOAD_PPS( (void const *) p2);
//					_MM512_STORE_PPS( (void *) p2, _reg1);
//					_MM512_STORE_PPS( (void *) p1, _reg2);
//				}
//			}
//		}
//	}
//	return;
//}

void call_slaswp(magma_mic_slaswp_param slaswp_param) {
	magma_int_t n, k1, k2, lda, incx;
	float *d_A;
	magma_int_t *d_ipiv;

	n		=	slaswp_param.n;
	d_A		=	(float *) slaswp_param.d_A;
	lda		=	slaswp_param.lda;
	k1		=	slaswp_param.k1;
	k2		=	slaswp_param.k2;
	d_ipiv		=	(magma_int_t *) slaswp_param.d_ipiv;
	incx		=	slaswp_param.incx;

	//printf("n: %d lda: %d k1: %d k2: %d incx: %d\n", n, lda, k1, k2, incx);
	slaswp(&n, (float *)d_A, &lda, &k1, &k2, (magma_int_t *)d_ipiv, &incx);
	return;
}
