/*
 *   -- micMAGMA (version 0.1) --
 *      Univ. of Tennessee, Knoxville
 *      Univ. of California, Berkeley
 *      Univ. of Colorado, Denver
 *      April 2012
 *
 * @precisions normal z -> s d c
 */

#include <mkl.h>

#include <mkl_trans.h>
#include <scif.h>
#include <fcntl.h>
#include <offload.h>

//#ifdef __MIC__
//#define _M512_TYPED
//#include <lmmintrin.h>
//#endif
#include <immintrin.h>


#include "magma.h"
//#include "magma_z_bindings.h"
#include "magma_z_server.h"



void call_zgemm(magma_mic_zgemm_param zgemm_param) {
	magma_trans_t transa, transb;
	magma_int_t m, n, k, lda, ldb, ldc;
	magmaDoubleComplex alpha, beta, *a, *b, *c;
	
        transa = zgemm_param.transa;
        transb = zgemm_param.transb;
        m = zgemm_param.m;
        n = zgemm_param.n;
        k = zgemm_param.k;
        alpha = zgemm_param.alpha;
        a = (magmaDoubleComplex *)zgemm_param.a;
        lda = zgemm_param.lda;
        b = (magmaDoubleComplex *)zgemm_param.b;
        ldb = zgemm_param.ldb;
        beta = zgemm_param.beta;
        c = (magmaDoubleComplex *)zgemm_param.c;
        ldc = zgemm_param.ldc;

        //printf("ZGEMM: %c %c %d %d %d %f 0x%lx %d 0x%lx %d %f 0x%lx %d\n", transa, transb, m, n, k, alpha, (off_t)a, lda, (off_t)b, ldb, beta, (off_t)c, ldc);

	zgemm( lapack_const(transa), lapack_const(transb), 
	       &m, &n, &k, (const MKL_Complex16 *)&alpha, (MKL_Complex16 *)a, &lda, 
	       (MKL_Complex16 *)b, &ldb, (const MKL_Complex16 *)&beta, 
	       (MKL_Complex16 *)c, &ldc);

	return;
}

void call_zgemv(magma_mic_zgemv_param zgemv_param) {
  magma_trans_t transa;
  magma_int_t m, n, lda, incx, incy;
  magmaDoubleComplex alpha, beta, *a, *x, *y;

  transa = zgemv_param.transa;
  m      = zgemv_param.m;
  n      = zgemv_param.n;
  alpha  = zgemv_param.alpha;
  a      = (magmaDoubleComplex *)zgemv_param.a;
  lda    = zgemv_param.lda;
  x      = (magmaDoubleComplex *)zgemv_param.x;
  incx   = zgemv_param.incx;
  beta   = zgemv_param.beta;
  y      = (magmaDoubleComplex *)zgemv_param.y;
  incy   = zgemv_param.incy;

  //printf("ZGEMV: %c %c %d %d %d %f 0x%lx %d 0x%lx %d %f 0x%lx %d\n", transa, transb, m, n, k, alpha, (off_t)a, lda, (off_t)b, ldb, beta, (off_t)c, ldc);                                                                           

  zgemv( lapack_const(transa), &m, &n, (const MKL_Complex16 *)&alpha, 
	 (MKL_Complex16 *)a, &lda, (MKL_Complex16 *)x, &incx, 
	 (const MKL_Complex16 *)&beta, (MKL_Complex16 *)y, &incy);

  return;
}

void call_ztrsm(magma_mic_ztrsm_param ztrsm_param) {
        magma_side_t side;
	magma_uplo_t uplo;
	magma_trans_t transa;
	magma_diag_t diag;
	magma_int_t m, n, lda, ldb;
	magmaDoubleComplex alpha, *a, *b;
	
        side = ztrsm_param.side;
        uplo = ztrsm_param.uplo;
        transa = ztrsm_param.transa;
        diag = ztrsm_param.diag;
        m = ztrsm_param.m;
        n = ztrsm_param.n;
        alpha = ztrsm_param.alpha;
        a = (magmaDoubleComplex *)ztrsm_param.a;
        lda = ztrsm_param.lda;
        b = (magmaDoubleComplex *)ztrsm_param.b;
        ldb = ztrsm_param.ldb;

        //printf("ZTRSM: %c %c %c %c %d %d %f 0x%lx %d 0x%lx %d\n", side, uplo, transa, diag, m, n, alpha, (off_t)a, lda, (off_t)b, ldb);
	ztrsm( lapack_const(side), lapack_const(uplo), lapack_const(transa), lapack_const(diag),
	       &m, &n, (const MKL_Complex16 *)&alpha, (MKL_Complex16 *)a, &lda, 
	       (MKL_Complex16 *)b, &ldb);

	return;
}


void call_ztrmm(magma_mic_ztrmm_param ztrmm_param) {
        magma_side_t side;
	magma_uplo_t uplo;
	magma_trans_t transa;
	magma_diag_t diag;
        magma_int_t m, n, lda, ldb;
        magmaDoubleComplex alpha, *a, *b;

        side = ztrmm_param.side;
        uplo = ztrmm_param.uplo;
        transa = ztrmm_param.transa;
        diag = ztrmm_param.diag;
        m = ztrmm_param.m;
        n = ztrmm_param.n;
        alpha = ztrmm_param.alpha;
        a = (magmaDoubleComplex *)ztrmm_param.a;
        lda = ztrmm_param.lda;
        b = (magmaDoubleComplex *)ztrmm_param.b;
        ldb = ztrmm_param.ldb;

        //printf("ZTRMM: %c %c %c %c %d %d %f 0x%lx %d 0x%lx %d\n", side, uplo, transa, diag, m, n, alpha, (off_t)a, lda, (off_t)b, ldb);
	ztrmm(lapack_const(side), lapack_const(uplo), lapack_const(transa), lapack_const(diag),
	      &m, &n, (const MKL_Complex16 *)&alpha, (MKL_Complex16 *)a, &lda, 
	      (MKL_Complex16 *)b, &ldb);

	return;
}
        


void call_zherk(magma_mic_zherk_param zherk_param) {
        magma_uplo_t uplo;
	magma_trans_t trans;
	magma_int_t n, k, lda, ldc;
	double alpha, beta;
	magmaDoubleComplex *a, *c;
	
        uplo = zherk_param.uplo;
        trans = zherk_param.trans;
        n = zherk_param.n;
        k = zherk_param.k;
        alpha = zherk_param.alpha;
        a = (magmaDoubleComplex *)zherk_param.a;
        lda = zherk_param.lda;
        beta = zherk_param.beta;
        c = (magmaDoubleComplex *)zherk_param.c;
        ldc = zherk_param.ldc;

        //printf("ZHERK: %c %c %d %d %f 0x%lx %d %f 0x%lx %d\n", uplo, trans, n, k, alpha, (off_t)a, lda, beta, (off_t)c, ldc);

	zherk(lapack_const(uplo), lapack_const(trans), 
	      &n, &k, &alpha, (MKL_Complex16 *)a, &lda, 
	      &beta, (MKL_Complex16 *)c, &ldc);

	return;
}


void call_zlarft(magma_mic_zlarft_param zlarft_param) {
	int direct, storev;
	magma_int_t n, k, ldv, ldt; 
	magmaDoubleComplex *v, *tau, *t;
 
        direct = zlarft_param.direct;
        storev = zlarft_param.storev;
        n = zlarft_param.n;
        k = zlarft_param.k;
        v = (magmaDoubleComplex *)zlarft_param.v;
        ldv = zlarft_param.ldv;
        tau = (magmaDoubleComplex *)zlarft_param.tau;
        t = (magmaDoubleComplex *)zlarft_param.t;
        ldt = zlarft_param.ldt;

	//printf("ZLARFT: %c %c %d %d 0x%lx %d ox%lx 0x%lx %d\n", direct, storev, n, k, (off_t)v, ldv, (off_t)tau, (off_t)t, ldt);

	zlarft(lapack_const(direct), lapack_const(storev),
	       &n, &k, (MKL_Complex16 *)v, &ldv, 
	       (MKL_Complex16 *)tau, (MKL_Complex16 *)t, &ldt);
	
	return;
}


void call_zlarfb(magma_mic_zlarfb_param zlarfb_param) {
	int side, trans, direct, storev;
	magma_int_t m, n, k, ldv, ldt, ldc, ldwork;
	magmaDoubleComplex *v, *t, *c, *work;

        side = zlarfb_param.side;
        trans = zlarfb_param.trans;
        direct = zlarfb_param.direct;
        storev = zlarfb_param.storev;
        m = zlarfb_param.m;
        n = zlarfb_param.n;
        k = zlarfb_param.k;
        v = (magmaDoubleComplex *)zlarfb_param.v;
        ldv = zlarfb_param.ldv;
        t = (magmaDoubleComplex *)zlarfb_param.t;
        ldt = zlarfb_param.ldt;
        c = (magmaDoubleComplex *)zlarfb_param.c;
        ldc = zlarfb_param.ldc;
        work = (magmaDoubleComplex *)zlarfb_param.work;
        ldwork = zlarfb_param.ldwork;

        //printf("ZLARFB: %c %c %c %c %d %d %d 0x%lx %d ox%lx %d 0x%lx %d 0x%lx %d\n", side, trans, direct, storev, m, n, k, (off_t)v, ldv, (off_t)t, ldt, (off_t)c, ldc, (off_t)work, ldwork); 
	zlarfb(lapack_const(side), lapack_const(trans), lapack_const(direct), lapack_const(storev),
			&m, &n, &k, (MKL_Complex16 *)v, &ldv, (MKL_Complex16 *)t, &ldt, 
			(MKL_Complex16 *)c, &ldc, (MKL_Complex16 *)work, &ldwork);

	/*
	magmaDoubleComplex c_zero    = MAGMA_Z_MAKE(  0.0, 0.0 );
	magmaDoubleComplex c_one     = MAGMA_Z_MAKE(  1.0, 0.0 );
	magmaDoubleComplex c_neg_one = MAGMA_Z_MAKE( -1.0, 0.0 );

	if (m <= 0 || n <= 0) {
		return;
	}

	magma_int_t transt;
	if (trans == MagmaNoTrans) 
		transt = MagmaConjTrans;
	else
		transt = MagmaNoTrans;

	if ( side  == MagmaLeft ) {

		if ( storev == MagmaColumnwise ) 
		{
			zgemm( lapack_const(MagmaConjTrans), lapack_const(MagmaNoTrans),
					&n, &k, &m,
					&c_one,  (MKL_Complex16 *)c,    &ldc,
					(MKL_Complex16 *)v,    &ldv,
					&c_zero, (MKL_Complex16 *)work, &ldwork);

			if (direct == MagmaForward)
				ztrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
						&n, &k, 
						&c_one, (MKL_Complex16 *)t,    &ldt, 
						(MKL_Complex16 *)work, &ldwork);
			else
				ztrmm( lapack_const(MagmaRight), lapack_const(MagmaLower), lapack_const(transt), lapack_const(MagmaNonUnit),
						&n, &k, 
						&c_one, (MKL_Complex16 *)t,    &ldt, 
						(MKL_Complex16 *)work, &ldwork);

			zgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaConjTrans), 
					&m, &n, &k, 
					&c_neg_one, (MKL_Complex16 *)v,    &ldv,
					(MKL_Complex16 *)work, &ldwork, 
					&c_one,     (MKL_Complex16 *)c,    &ldc);
		}
		else {
			zgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaConjTrans), 
					&m, &k, &n, 
					&c_one,  (MKL_Complex16 *)c,    &ldc,
					(MKL_Complex16 *)v,    &ldv, 
					&c_zero, (MKL_Complex16 *)work, &ldwork);

			ztrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
					&m, &k, 
					&c_one, (MKL_Complex16 *)t,    &ldt, 
					(MKL_Complex16 *)work, &ldwork);

			zgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans), 
					&m, &n, &k, 
					&c_neg_one, (MKL_Complex16 *)work, &ldwork,
					(MKL_Complex16 *)v,    &ldv,
					&c_one,     (MKL_Complex16 *)c,    &ldc);
		}
	}

	else {

		if ( storev == MagmaColumnwise ) {
			zgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans),
					&m, &k, &n,
					&c_one,  (MKL_Complex16 *)c,    &ldc,
					(MKL_Complex16 *)v,    &ldv,
					&c_zero, (MKL_Complex16 *)work, &ldwork);
			// ??? ldwork replaced by k for case n < k

			if (direct == MagmaForward)
				ztrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
						&m, &k,
						&c_one, (MKL_Complex16 *)t,    &ldt,
						(MKL_Complex16 *)work, &ldwork);
			else
				ztrmm( lapack_const(MagmaRight), lapack_const(MagmaLower), lapack_const(transt), lapack_const(MagmaNonUnit),
						&m, &k,
						&c_one, (MKL_Complex16 *)t,    &ldt,
						(MKL_Complex16 *)work, &ldwork);

			zgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaConjTrans),
					&m, &n, &k,
					&c_neg_one, (MKL_Complex16 *)work, &ldwork, 
					(MKL_Complex16 *)v,    &ldv,
					&c_one,     (MKL_Complex16 *)c,    &ldc);
		}
		else {
			zgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaConjTrans),
					&m, &k, &n,
					&c_one,  (MKL_Complex16 *)c,    &ldc,
					(MKL_Complex16 *)v,    &ldv,
					&c_zero, (MKL_Complex16 *)work, &ldwork);

			ztrmm( lapack_const(MagmaRight), lapack_const(MagmaUpper), lapack_const(transt), lapack_const(MagmaNonUnit),
					&m, &k,
					&c_one, (MKL_Complex16 *)t,    &ldt,
					(MKL_Complex16 *)work, &ldwork);

			zgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans),
					&m, &n, &k,
					&c_neg_one, (MKL_Complex16 *)work, &ldwork,
					(MKL_Complex16 *)v,    &ldv,
					&c_one,     (MKL_Complex16 *)c,    &ldc);
		}
	}*/

	return;
}


void call_ztranspose(magma_mic_ztranspose_param ztranspose_param)
{
	/******************************************************
	 * _mm512_load_pd -- for loading
	 * _mm512_i32lo[ext]scatter_pd
	 * _mm512_i32loscatter_pd is the simplified version of ...ext...
	 * they do the same thing, the latter needs less arguments.
	 * **************************************************/
	//	printf("entered stranpose\n");
	//	fflush(stdout);
	double *odata, *idata;
	int ldo,ldi, m, n;
	int num_threads;
	int tid;
	int index[16], i;
	__M512Z _reg;
	__m512i _index;
	double *my_head, *loop_head, *dest_head;
	int num_cols, num_cols_g, col, iter, num_iter, num_rows_g, num_rows;

	odata	=	(double *) ztranspose_param.odata;
	ldo		=	ztranspose_param.ldo;
	idata	=	(double *) ztranspose_param.idata;
	ldi		=	ztranspose_param.ldi;
	m		=	ztranspose_param.m;
	n		=	ztranspose_param.n;

	if (m == 0 || n == 0) {
		return;
	}
	if (m <= n) {
		for (i = 0; i < 16; i++) {
			index[i] = i*ZREG_WIDTH*ldo;
		}
		// 62 cores, 4 hardware threads per core. Leave out one core for the OS
		// (62 - 1)*4 = 244

		num_threads = 244;
		num_cols_g =	(n + num_threads - 1)/num_threads; 
		num_iter = (int) m*1.0/ZREG_WIDTH;
		_index = _mm512_set_epi32(index[15], index[14], index[13], index[12],
				index[11], index[10], index[9], index[8], index[7], index[6],
				index[5], index[4], index[3], index[2], index[1], index[0]);

#pragma omp parallel num_threads(num_threads) private(tid, col, num_cols, iter,\
		my_head, loop_head, dest_head, _reg)
		{
			tid = omp_get_thread_num();
			num_cols = num_cols_g;
			if (tid == (num_threads - 1) && (n%num_cols_g != 0)) { 
				num_cols = n%num_cols_g;
			}
			// starting point of my section, constant for each thread
			my_head = idata + ldi*(tid*num_cols_g);
			for (iter = 0; iter < num_iter; iter++) { // loop inside each column
				for (col = 0; col < num_cols; col++) { // loop over each column
					if (tid*num_cols_g + col < n) {
						// starting point of the inner loop
						loop_head = my_head + col*ldi + iter*ZREG_WIDTH;
						dest_head = odata + (iter*ZREG_WIDTH)*ldo + (tid*num_cols_g + col);
						// load 8 double precision floating point nubmers from address
						// mt into a float64 vector.
						_reg = _MM512_LOAD_PPZ( (void const *)loop_head);
						//					_mm512_load_pd( (void const *)loop_head);

						// this intrinsic needs int32 indices, but a 512 register
						// contains 16 int32 integers, and there is only one intrinsic
						// to set all 16 positoins of the register. So _index contains
						// 16 int32 integers.
						_MM512_I32LOEXTSCATTER_PPZ((void *) dest_head, _index, _reg,
								_MM_DOWNCONV_PZ_NONE, 1, _MM_HINT_NONE);
					}
				}
			}	
		}
	} else {
		// init index[], it's for scattering the entries to the dest matrix
		for (i = 0; i < 16; i++) {
			index[i] = i*ZREG_WIDTH*ldo;
		}

		// 62 cores, 4 hardware threads per core. Leave out one core for the OS
		// (62 - 1)*4 = 244
		num_threads = 244;
		// number of rows in each group
		num_rows_g = (m/ZREG_WIDTH + num_threads - 1)/num_threads*ZREG_WIDTH;
		_index = _mm512_set_epi32(index[15], index[14], index[13], index[12],
				index[11], index[10], index[9], index[8], index[7], index[6],
				index[5], index[4], index[3], index[2], index[1], index[0]);

#pragma omp parallel num_threads(num_threads) private(tid, col, num_rows, iter,\
		loop_head, dest_head, _reg, num_iter)
		{
			tid = omp_get_thread_num();
			// compute the number of rows I need to work on
			num_rows = num_rows_g;
			if (tid == (num_threads - 1) && (m%num_rows_g != 0)) {
				num_rows = m%num_rows_g;
			}
			num_iter = (int) num_rows*1.0/ZREG_WIDTH;

			for (iter = 0; iter < num_iter; iter++) {
				if (tid*num_rows_g + iter*ZREG_WIDTH < m) {
					for (col = 0; col < n; col++) {
						loop_head = idata + (tid*num_rows_g + iter*ZREG_WIDTH) + ldi*col;
						dest_head = odata + ldo*(tid*num_rows_g + iter*ZREG_WIDTH) + col;

						_reg = _MM512_LOAD_PPZ((void const *) loop_head);
						_MM512_I32LOEXTSCATTER_PPZ((void *) dest_head, _index, _reg,
								_MM_DOWNCONV_PZ_NONE, 1, _MM_HINT_NONE);
					}
				}
			}

		}
	}
	return;
}

void call_ztranspose_mkl(magma_mic_ztranspose_param ztranspose_param)
{
	magmaDoubleComplex *odata, *idata;
	int ldo,ldi, m, n;
	int num_threads;
	int tid;
	int index[16], i;
	//  _M512 _reg;
	//  _M512I _index;
	magmaDoubleComplex *my_head, *loop_head, *dest_head;
	int num_cols, num_cols_g, col, iter, num_iter;

	odata =(magmaDoubleComplex *) ztranspose_param.odata;
	ldo   =ztranspose_param.ldo;
	idata =(magmaDoubleComplex *) ztranspose_param.idata;
	ldi   = ztranspose_param.ldi;
	m     = ztranspose_param.m;
	n     = ztranspose_param.n;
	magmaDoubleComplex alpha = MAGMA_Z_MAKE(  1.0, 0.0 );

	mkl_zomatcopy('C', 'T', m, n, *((MKL_Complex16 *)(&alpha)), (MKL_Complex16 *)idata, ldi, 
			(MKL_Complex16 *)odata, ldo);

	return;
}

//void call_zlaswp(magma_mic_zlaswp_param zlaswp_param) {
	/*
	 * takes 0-based ipiv
	 * In magma semantics, d_AT is always the starting point of the matrix, it
	 * doesn't change.
	 */
	//	fprintf(stdout, "entering zlaswp\n");
//	double *d_AT;
//	int lda, nb, ind;
//	int *d_ipiv;

//	d_AT		=	(double *) zlaswp_param.d_AT;
//	lda			=	zlaswp_param.lda;
//	d_ipiv		=	(int *) zlaswp_param.d_ipiv;
//	nb			=	zlaswp_param.nb;
//	ind			=	zlaswp_param.ind;

//	if (nb == 0) {
//		return;
//	}

//	int my_i, my_j;
//	int tid;
//	double *p1, *p2, tmp, *head;
//	int  *ipiv_tmp;
//	int num_threads;
//	int blk, num_blk;

//	__M512Z _reg1, _reg2;

//	ipiv_tmp = d_ipiv;
	// (62 - 1)*4 = 244
//	num_threads		= 244;
	// the number of 8-column blocks in each section. Every thread takes one
	// section.
//	num_blk			= (lda/ZREG_WIDTH + num_threads - 1)/num_threads;
	//	printf("num_blk = %d\n", num_blk);
//#pragma omp parallel num_threads(num_threads) private(tid, p1, p2, tmp, head,\
//		my_i, my_j, blk, _reg1, _reg2)
//	{
		/* 
		 * num_reps		the number of 512 vectors this thread owns
		 */
//		tid = omp_get_thread_num();
//		for (my_i = 0; my_i < nb; my_i++) { // only swap nb rows
			// my_j is the destination row index
			// convert 1-based to 0 based ipiv
//			my_j = d_ipiv[my_i + ind] - 1 + ind; 
//			if (my_i + ind == my_j) continue;
			// in my section, work on the 8-column blocks one by one
//			for (blk = 0; blk < num_blk; blk++) { // loop in a row
//				head = d_AT + tid*(num_blk*ZREG_WIDTH) + blk*ZREG_WIDTH;
				// check the end of this thread is in range
//				if (tid*(num_blk*ZREG_WIDTH) + blk*ZREG_WIDTH < lda) { 
					// p1 -- the source element
					// p2 -- the dest element
//					p1 = head + (my_i + ind)*lda;
//					p2 = head + my_j*lda;
//					_reg1 = _MM512_LOAD_PPZ( (void const *) p1);
//					_reg2 = _MM512_LOAD_PPZ( (void const *) p2);
//					_MM512_STORE_PPZ( (void *) p2, _reg1);
//					_MM512_STORE_PPZ( (void *) p1, _reg2);
//				}
//			}
//		}
//	}
//	return;
//}

void call_zlaswp(magma_mic_zlaswp_param zlaswp_param) {
	magma_int_t n, k1, k2, lda, incx;
	magmaDoubleComplex *d_A;
	magma_int_t *d_ipiv;

	n		=	zlaswp_param.n;
	d_A		=	(magmaDoubleComplex *) zlaswp_param.d_A;
	lda		=	zlaswp_param.lda;
	k1		=	zlaswp_param.k1;
	k2		=	zlaswp_param.k2;
	d_ipiv		=	(magma_int_t *) zlaswp_param.d_ipiv;
	incx		=	zlaswp_param.incx;

	//printf("n: %d lda: %d k1: %d k2: %d incx: %d\n", n, lda, k1, k2, incx);
	zlaswp(&n, (MKL_Complex16 *)d_A, &lda, &k1, &k2, (magma_int_t *)d_ipiv, &incx);
	return;
}
